from libc.stdio cimport FILE
from .c_sundials cimport *

cdef extern from "sunmatrix/sunmatrix_dense.h":

    cdef struct _SUNMatrixContent_Dense:
        sunindextype M
        sunindextype N
        sunrealtype *data
        sunindextype ldata
        sunrealtype **cols

    ctypedef _SUNMatrixContent_Dense *SUNMatrixContent_Dense

    SUNMatrix SUNDenseMatrix(sunindextype M, sunindextype N, SUNContext sunctx)
    void SUNDenseMatrix_Print(SUNMatrix A, FILE* outfile)
    sunindextype SUNDenseMatrix_Rows(SUNMatrix A)
    sunindextype SUNDenseMatrix_Columns(SUNMatrix A)
    sunindextype SUNDenseMatrix_LData(SUNMatrix A)
    sunrealtype* SUNDenseMatrix_Data(SUNMatrix A)
    sunrealtype** SUNDenseMatrix_Cols(SUNMatrix A)
    sunrealtype* SUNDenseMatrix_Column(SUNMatrix A, sunindextype j)
    SUNMatrix_ID SUNMatGetID_Dense(SUNMatrix A)
    SUNMatrix SUNMatClone_Dense(SUNMatrix A)
    void SUNMatDestroy_Dense(SUNMatrix A)
    int SUNMatZero_Dense(SUNMatrix A)
    int SUNMatCopy_Dense(SUNMatrix A, SUNMatrix B)
    int SUNMatScaleAdd_Dense(sunrealtype c, SUNMatrix A, SUNMatrix B)
    int SUNMatScaleAddI_Dense(sunrealtype c, SUNMatrix A)
    int SUNMatMatvec_Dense(SUNMatrix A, N_Vector x, N_Vector y)
    int SUNMatSpace_Dense(SUNMatrix A, long int *lenrw, long int *leniw)


cdef extern from "sunmatrix/sunmatrix_band.h":

    cdef struct _SUNMatrixContent_Band:
        sunindextype M
        sunindextype N
        sunindextype ldim
        sunindextype mu
        sunindextype ml
        sunindextype s_mu
        sunrealtype *data
        sunindextype ldata
        sunrealtype **cols

    ctypedef _SUNMatrixContent_Band *SUNMatrixContent_Band

    SUNMatrix SUNBandMatrix(sunindextype N, sunindextype mu,
                                            sunindextype ml, SUNContext sunctx)

    SUNMatrix SUNBandMatrixStorage(sunindextype N, sunindextype mu,
                                   sunindextype ml, sunindextype smu, SUNContext sunctx)

    void SUNBandMatrix_Print(SUNMatrix A, FILE* outfile)

    sunindextype SUNBandMatrix_Rows(SUNMatrix A)
    sunindextype SUNBandMatrix_Columns(SUNMatrix A)
    sunindextype SUNBandMatrix_LowerBandwidth(SUNMatrix A)
    sunindextype SUNBandMatrix_UpperBandwidth(SUNMatrix A)
    sunindextype SUNBandMatrix_StoredUpperBandwidth(SUNMatrix A)
    sunindextype SUNBandMatrix_LDim(SUNMatrix A)
    sunrealtype* SUNBandMatrix_Data(SUNMatrix A)
    sunrealtype** SUNBandMatrix_Cols(SUNMatrix A)
    sunrealtype* SUNBandMatrix_Column(SUNMatrix A, sunindextype j)

    SUNMatrix_ID SUNMatGetID_Band(SUNMatrix A)
    SUNMatrix SUNMatClone_Band(SUNMatrix A)
    void SUNMatDestroy_Band(SUNMatrix A)
    int SUNMatZero_Band(SUNMatrix A)
    int SUNMatCopy_Band(SUNMatrix A, SUNMatrix B)
    int SUNMatScaleAdd_Band(sunrealtype c, SUNMatrix A, SUNMatrix B)
    int SUNMatScaleAddI_Band(sunrealtype c, SUNMatrix A)
    int SUNMatMatvec_Band(SUNMatrix A, N_Vector x, N_Vector y)
    int SUNMatSpace_Band(SUNMatrix A, long int *lenrw, long int *leniw)


cdef extern from "sunmatrix/sunmatrix_sparse.h":

    cdef struct _SUNMatrixContent_Sparse:
        sunindextype M
        sunindextype N
        sunindextype NNZ
        sunindextype NP
        sunrealtype *data
        int sparsetype
        sunindextype *indexvals
        sunindextype *indexptrs
        # CSC indices
        sunindextype **rowvals
        sunindextype **colptrs
        # CSR indices
        sunindextype **colvals
        sunindextype **rowptrs

    ctypedef _SUNMatrixContent_Sparse *SUNMatrixContent_Sparse


    SUNMatrix SUNSparseMatrix(sunindextype M, sunindextype N,
                              sunindextype NNZ, int sparsetype)
    SUNMatrix SUNSparseFromDenseMatrix(SUNMatrix A, sunrealtype droptol,
                                       int sparsetype)
    SUNMatrix SUNSparseFromBandMatrix(SUNMatrix A, sunrealtype droptol,
                                      int sparsetype)
    int SUNSparseMatrix_Realloc(SUNMatrix A)
    void SUNSparseMatrix_Print(SUNMatrix A, FILE* outfile)

    sunindextype SUNSparseMatrix_Rows(SUNMatrix A)
    sunindextype SUNSparseMatrix_Columns(SUNMatrix A)
    sunindextype SUNSparseMatrix_NNZ(SUNMatrix A)
    sunindextype SUNSparseMatrix_NP(SUNMatrix A)
    int SUNSparseMatrix_SparseType(SUNMatrix A)
    sunrealtype* SUNSparseMatrix_Data(SUNMatrix A)
    sunindextype* SUNSparseMatrix_IndexValues(SUNMatrix A)
    sunindextype* SUNSparseMatrix_IndexPointers(SUNMatrix A)

    SUNMatrix_ID SUNMatGetID_Sparse(SUNMatrix A)
    SUNMatrix SUNMatClone_Sparse(SUNMatrix A)
    void SUNMatDestroy_Sparse(SUNMatrix A)
    int SUNMatZero_Sparse(SUNMatrix A)
    int SUNMatCopy_Sparse(SUNMatrix A, SUNMatrix B)
    int SUNMatScaleAdd_Sparse(sunrealtype c, SUNMatrix A, SUNMatrix B)
    int SUNMatScaleAddI_Sparse(sunrealtype c, SUNMatrix A)
    int SUNMatMatvec_Sparse(SUNMatrix A, N_Vector x, N_Vector y)
    int SUNMatSpace_Sparse(SUNMatrix A, long int *lenrw, long int *leniw)



#defining NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION

cimport numpy as np
from .c_sundials cimport N_Vector, sunrealtype, SUNContext
from .common_defs cimport DTYPE_t, INDEX_TYPE_t

cdef class CV_RhsFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] ydot,
                       object userdata = *) except? -1

cdef class CV_WrapRhsFunction(CV_RhsFunction):
    cdef public object _rhsfn
    cdef public int with_userdata
    cpdef set_rhsfn(self, object rhsfn)

cdef class CV_RootFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] g,
                       object userdata = *) except? -1

cdef class CV_WrapRootFunction(CV_RootFunction):
    cdef object _rootfn
    cdef int with_userdata
    cpdef set_rootfn(self, object rootfn)

cdef class CV_JacRhsFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] fy,
                       np.ndarray[DTYPE_t, ndim=2] J) except? -1

cdef class CV_WrapJacRhsFunction(CV_JacRhsFunction):
    cdef public object _jacfn
    cdef int with_userdata
    cpdef set_jacfn(self, object jacfn)

cdef class CV_PrecSetupFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       bint jok,
                       object jcurPtr,
                       DTYPE_t gamma,
                       object userdata = *) except? -1

cdef class CV_WrapPrecSetupFunction(CV_PrecSetupFunction):
    cdef object _prec_setupfn
    cdef int with_userdata
    cpdef set_prec_setupfn(self, object prec_setupfn)

cdef class CV_PrecSolveFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] r,
                       np.ndarray[DTYPE_t, ndim=1] z,
                       DTYPE_t gamma,
                       DTYPE_t delta,
                       int lr,
                       object userdata = *) except? -1

cdef class CV_WrapPrecSolveFunction(CV_PrecSolveFunction):
    cdef object _prec_solvefn
    cdef int with_userdata
    cpdef set_prec_solvefn(self, object prec_solvefn)


cdef class CV_JacTimesVecFunction:
    cpdef int evaluate(self,
                       np.ndarray[DTYPE_t, ndim=1] v,
                       np.ndarray[DTYPE_t, ndim=1] Jv,
                       DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       object userdata = *) except? -1

cdef class CV_WrapJacTimesVecFunction(CV_JacTimesVecFunction):
    cdef object _jac_times_vecfn
    cdef int with_userdata
    cpdef set_jac_times_vecfn(self, object jac_times_vecfn)

cdef class CV_JacTimesSetupFunction:
    cpdef int evaluate(self,
                       DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] fy,
                       object userdata = *) except? -1

cdef class CV_WrapJacTimesSetupFunction(CV_JacTimesSetupFunction):
    cdef object _jac_times_setupfn
    cdef int with_userdata
    cpdef set_jac_times_setupfn(self, object jac_times_setupfn)

cdef class CV_ContinuationFunction:
    cdef object _fn
    cpdef int evaluate(self, DTYPE_t t, np.ndarray[DTYPE_t, ndim=1] y,
                       CVODE solver)

cdef class CV_ErrHandler:
    cpdef evaluate(self,
                   int error_code,
                   bytes module,
                   bytes function,
                   bytes msg,
                   object user_data = *)

cdef class CV_WrapErrHandler(CV_ErrHandler):
    cdef object _err_handler
    cdef int with_userdata
    cpdef set_err_handler(self, object err_handler)


cdef class CV_data:
    cdef np.ndarray yy_tmp, yp_tmp, jac_tmp, g_tmp, r_tmp, z_tmp
    cdef CV_RhsFunction rfn
    cdef CV_JacRhsFunction jac
    cdef CV_RootFunction rootfn
    cdef CV_PrecSolveFunction prec_solvefn
    cdef CV_PrecSetupFunction prec_setupfn
    cdef CV_JacTimesVecFunction jac_times_vecfn
    cdef CV_JacTimesSetupFunction jac_times_setupfn
    cdef bint parallel_implementation
    cdef object user_data
    cdef CV_ErrHandler err_handler
    cdef object err_user_data

cdef class CVODE:
    cdef N_Vector atol
    cdef void* _cv_mem
    cdef SUNContext sunctx
    cdef dict options
    cdef bint parallel_implementation, initialized, _old_api, _step_compute, _validate_flags
    cdef CV_data aux_data

    cdef INDEX_TYPE_t N #problem size, i.e. len(y0) = N
    cdef N_Vector y0, y, yp # for 'step' method
    cdef list t_roots
    cdef list y_roots
    cdef list t_tstop
    cdef list y_tstop

    cdef int verbosity

    # Functions
    cpdef _create_suncontext(self)
    cpdef _init_step(self, DTYPE_t t0, np.ndarray[DTYPE_t, ndim=1] y0)
    cpdef _reinit_IC(self, DTYPE_t t0, np.ndarray[DTYPE_t, ndim=1] y0)

    cpdef _solve(self, np.ndarray[DTYPE_t, ndim=1] tspan,
                       np.ndarray[DTYPE_t, ndim=1] y0)
    cpdef _set_runtime_changeable_options(self, object options,
                                          bint supress_supported_check=*)



from .c_sundials cimport *
from .c_sunmatrix cimport *

include "sundials_config.pxi"


cdef extern from "sunnonlinsol/sunnonlinsol_newton.h":

    struct _SUNNonlinearSolverContent_Newton:
    
      SUNNonlinSolSysFn      Sys
      SUNNonlinSolLSetupFn   LSetup
      SUNNonlinSolLSolveFn   LSolve
      SUNNonlinSolConvTestFn CTest
    
      N_Vector    delta
      sunbooleantype jcur
      int         curiter
      int         maxiters
      long int    niters
      long int    nconvfails
      void*       ctest_data; 
    
    ctypedef _SUNNonlinearSolverContent_Newton *SUNNonlinearSolverContent_Newton;

    SUNNonlinearSolver SUNNonlinSol_Newton(N_Vector y, SUNContext sunctx)
    SUNNonlinearSolver SUNNonlinSol_NewtonSens(int count, N_Vector y, SUNContext sunctx)

    SUNNonlinearSolver_Type SUNNonlinSolGetType_Newton(SUNNonlinearSolver NLS)

    int SUNNonlinSolInitialize_Newton(SUNNonlinearSolver NLS)
    int SUNNonlinSolSolve_Newton(SUNNonlinearSolver NLS,  N_Vector y0, 
                                 N_Vector y, N_Vector w, sunrealtype tol,
                                 sunbooleantype callLSetup, void *mem)
    int SUNNonlinSolFree_Newton(SUNNonlinearSolver NLS)

    int SUNNonlinSolSetSysFn_Newton(SUNNonlinearSolver NLS,
                                    SUNNonlinSolSysFn SysFn);
    int SUNNonlinSolSetLSetupFn_Newton(SUNNonlinearSolver NLS,
                                       SUNNonlinSolLSetupFn LSetupFn);
    int SUNNonlinSolSetLSolveFn_Newton(SUNNonlinearSolver NLS,
                                       SUNNonlinSolLSolveFn LSolveFn);
    int SUNNonlinSolSetConvTestFn_Newton(SUNNonlinearSolver NLS,
                                         SUNNonlinSolConvTestFn CTestFn,
                                         void* ctest_data);
    int SUNNonlinSolSetMaxIters_Newton(SUNNonlinearSolver NLS,
                                       int maxiters);

    int SUNNonlinSolGetNumIters_Newton(SUNNonlinearSolver NLS,
                                       long int *niters);
    int SUNNonlinSolGetCurIter_Newton(SUNNonlinearSolver NLS, int *iter);
    int SUNNonlinSolGetNumConvFails_Newton(SUNNonlinearSolver NLS,
                                           long int *nconvfails);
    int SUNNonlinSolGetSysFn_Newton(SUNNonlinearSolver NLS,
                                    SUNNonlinSolSysFn *SysFn);


cdef extern from "sunnonlinsol/sunnonlinsol_fixedpoint.h":
    
    struct _SUNNonlinearSolverContent_FixedPoint:

        SUNNonlinSolSysFn      Sys
        SUNNonlinSolConvTestFn CTest

        int       m
        int      *imap
        sunrealtype *R
        sunrealtype *gamma
        sunrealtype *cvals
        N_Vector *df
        N_Vector *dg
        N_Vector *q
        N_Vector *Xvecs
        N_Vector  yprev
        N_Vector  gy
        N_Vector  fold
        N_Vector  gold
        N_Vector  delta
        int       curiter
        int       maxiters
        long int  niters
        long int  nconvfails
        void*     ctest_data

    ctypedef _SUNNonlinearSolverContent_FixedPoint *SUNNonlinearSolverContent_FixedPoint

    SUNNonlinearSolver SUNNonlinSol_FixedPoint(N_Vector y, int m, SUNContext sunctx)
    SUNNonlinearSolver SUNNonlinSol_FixedPointSens(int count, N_Vector y, int m, SUNContext sunctx)

    SUNNonlinearSolver_Type SUNNonlinSolGetType_FixedPoint(SUNNonlinearSolver NLS)

    int SUNNonlinSolInitialize_FixedPoint(SUNNonlinearSolver NLS)
    int SUNNonlinSolSolve_FixedPoint(SUNNonlinearSolver NLS,
                                     N_Vector y0, N_Vector y,
                                     N_Vector w, sunrealtype tol,
                                     sunbooleantype callSetup, void *mem)
    int SUNNonlinSolFree_FixedPoint(SUNNonlinearSolver NLS)

    int SUNNonlinSolSetSysFn_FixedPoint(SUNNonlinearSolver NLS,
                                        SUNNonlinSolSysFn SysFn);
    int SUNNonlinSolSetConvTestFn_FixedPoint(SUNNonlinearSolver NLS,
                                             SUNNonlinSolConvTestFn CTestFn,
                                             void* ctest_data);
    int SUNNonlinSolSetMaxIters_FixedPoint(SUNNonlinearSolver NLS,
                                           int maxiters);

    int SUNNonlinSolGetNumIters_FixedPoint(SUNNonlinearSolver NLS,
                                           long int *niters)
    int SUNNonlinSolGetCurIter_FixedPoint(SUNNonlinearSolver NLS, int *iter)
    int SUNNonlinSolGetNumConvFails_FixedPoint(SUNNonlinearSolver NLS,
                                               long int *nconvfails)
    int SUNNonlinSolGetSysFn_FixedPoint(SUNNonlinearSolver NLS,
                                        SUNNonlinSolSysFn *SysFn)


# We don't use PETSc SNES nonlinear solvers- "petscsnes.h" required then
#  as dependency, see https://www.mcs.anl.gov/petsc/documentation/installation.html 
# =============================================================================
# cdef extern from "sunnonlinsol_petscsnes.h":
#     
#     struct _SUNNonlinearSolverContent_PetscSNES:
#         
#         int sysfn_last_err
#         PetscErrorCode petsc_last_err
#         long int nconvfails
#         long int nni
#         void *imem
#         SNES snes
#         Vec r
#         N_Vector y, f
#         SUNNonlinSolSysFn Sys
# 
#     ctypedef _SUNNonlinearSolverContent_PetscSNES *SUNNonlinearSolverContent_PetscSNES
# 
#     SUNNonlinearSolver SUNNonlinSol_PetscSNES(N_Vector y, SNES snes, SUNContext sunctx)
#     SUNNonlinearSolver_Type SUNNonlinSolGetType_PetscSNES(SUNNonlinearSolver NLS)
# 
#     int SUNNonlinSolInitialize_PetscSNES(SUNNonlinearSolver NLS)
#     int SUNNonlinSolSolve_PetscSNES(SUNNonlinearSolver NLS,
#                                     N_Vector y0, N_Vector y,
#                                     N_Vector w, sunrealtype tol,
#                                     sunbooleantype callLSetup, void* mem)
# 
#     int SUNNonlinSolSetSysFn_PetscSNES(SUNNonlinearSolver NLS,
#                                        SUNNonlinSolSysFn SysFn)
# 
#     int SUNNonlinSolGetNumIters_PetscSNES(SUNNonlinearSolver NLS, long int* nni);
#     int SUNNonlinSolGetNumConvFails_PetscSNES(SUNNonlinearSolver NLS,
#                                               long int* nconvfails)
#     int SUNNonlinSolFree_PetscSNES(SUNNonlinearSolver NLS)
#     int SUNNonlinSolGetSNES_PetscSNES(SUNNonlinearSolver NLS, SNES* snes)
#     int SUNNonlinSolGetPetscError_PetscSNES(SUNNonlinearSolver NLS,
#                                             PetscErrorCode* err)
#     int SUNNonlinSolGetSysFn_PetscSNES(SUNNonlinearSolver NLS,
#                                        SUNNonlinSolSysFn* SysFn)
# 
# =============================================================================



cimport numpy as np
from .c_sundials cimport N_Vector, sunrealtype, SUNContext
from .common_defs cimport DTYPE_t

cdef class IDA_RhsFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] ydot,
                       np.ndarray[DTYPE_t, ndim=1] result,
                       object userdata = *) except? -1

cdef class IDA_WrapRhsFunction(IDA_RhsFunction):
    cdef object _resfn
    cdef int with_userdata
    cpdef set_resfn(self, object resfn)

cdef class IDA_RootFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] ydot,
                       np.ndarray[DTYPE_t, ndim=1] g,
                       object userdata = *) except? -1

cdef class IDA_WrapRootFunction(IDA_RootFunction):
    cdef object _rootfn
    cdef int with_userdata
    cpdef set_rootfn(self, object rootfn)

cdef class IDA_JacRhsFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] ydot,
                       np.ndarray[DTYPE_t, ndim=1] residual,
                       DTYPE_t cj,
                       np.ndarray[DTYPE_t, ndim=2] J,
                       object userdata = *) except? -1


cdef class IDA_WrapJacRhsFunction(IDA_JacRhsFunction):
    cdef object _jacfn
    cdef int with_userdata
    cpdef set_jacfn(self, object jacfn) 

cdef class IDA_PrecSetupFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] yp,
                       np.ndarray[DTYPE_t, ndim=1] rr,
                       DTYPE_t cj,
                       object userdata = *) except? -1

cdef class IDA_WrapPrecSetupFunction(IDA_PrecSetupFunction):
    cdef object _prec_setupfn
    cdef int with_userdata
    cpdef set_prec_setupfn(self, object prec_setupfn)

cdef class IDA_PrecSolveFunction:
    cpdef int evaluate(self, DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] yp,
                       np.ndarray[DTYPE_t, ndim=1] r,
                       np.ndarray[DTYPE_t, ndim=1] rvec,
                       np.ndarray[DTYPE_t, ndim=1] z,
                       DTYPE_t cj,
                       DTYPE_t delta,
                       object userdata = *) except? -1
                           
cdef class IDA_WrapPrecSolveFunction(IDA_PrecSolveFunction):
    cdef object _prec_solvefn
    cdef int with_userdata
    cpdef set_prec_solvefn(self, object prec_solvefn)

cdef class IDA_JacTimesVecFunction:
    cpdef int evaluate(self,
                       DTYPE_t t,
                       np.ndarray[DTYPE_t, ndim=1] yy,
                       np.ndarray[DTYPE_t, ndim=1] yp,
                       np.ndarray[DTYPE_t, ndim=1] rr,
                       np.ndarray[DTYPE_t, ndim=1] v,
                       np.ndarray[DTYPE_t, ndim=1] Jv,
                       DTYPE_t cj,
                       object userdata = *) except? -1

cdef class IDA_WrapJacTimesVecFunction(IDA_JacTimesVecFunction):
    cdef object _jac_times_vecfn
    cdef int with_userdata
    cpdef set_jac_times_vecfn(self, object jac_times_vecfn)

cdef class IDA_JacTimesSetupFunction:
    cpdef int evaluate(self,
                       DTYPE_t tt,
                       np.ndarray[DTYPE_t, ndim=1] yy,
                       np.ndarray[DTYPE_t, ndim=1] yp,
                       np.ndarray[DTYPE_t, ndim=1] rr,
                       DTYPE_t cj,
                       object userdata = *) except? -1

cdef class IDA_WrapJacTimesSetupFunction(IDA_JacTimesSetupFunction):
    cdef object _jac_times_setupfn
    cdef int with_userdata
    cpdef set_jac_times_setupfn(self, object jac_times_setupfn)

    cpdef int evaluate(self,
                       DTYPE_t tt,
                       np.ndarray[DTYPE_t, ndim=1] yy,
                       np.ndarray[DTYPE_t, ndim=1] yp,
                       np.ndarray[DTYPE_t, ndim=1] rr,
                       DTYPE_t cj,
                       object userdata = *) except? -1

cdef class IDA_ContinuationFunction:
    cdef object _fn
    cpdef int evaluate(self, DTYPE_t t, np.ndarray[DTYPE_t, ndim=1] y,
                       np.ndarray[DTYPE_t, ndim=1] ydot,
                       IDA solver)

cdef int _res(sunrealtype tt, N_Vector yy, N_Vector yp, N_Vector rr, void *self_obj)

cdef class IDA_ErrHandler:
    cpdef evaluate(self,
                   int error_code,
                   bytes module,
                   bytes function,
                   bytes msg,
                   object user_data = *)

cdef class IDA_WrapErrHandler(IDA_ErrHandler):
    cdef object _err_handler
    cdef int with_userdata
    cpdef set_err_handler(self, object err_handler)


cdef class IDA_data:
    cdef np.ndarray yy_tmp, yp_tmp, residual_tmp, jac_tmp
    cdef np.ndarray g_tmp, z_tmp, rvec_tmp, v_tmp
    cdef IDA_RhsFunction res
    cdef IDA_JacRhsFunction jac
    cdef IDA_RootFunction rootfn
    cdef IDA_PrecSetupFunction prec_setupfn
    cdef IDA_PrecSolveFunction prec_solvefn
    cdef IDA_JacTimesVecFunction jac_times_vecfn
    cdef IDA_JacTimesSetupFunction jac_times_setupfn
    cdef bint parallel_implementation
    cdef object user_data
    cdef IDA_ErrHandler err_handler
    cdef object err_user_data

cdef class IDA:
    cdef N_Vector atol

    cdef N_Vector y0, yp0, residual, y, yp
    cdef N_Vector dae_vars_id, constraints
    cdef long int N #problem size, i.e. len(y0) = N
    cdef list t_roots
    cdef list y_roots
    cdef list yp_roots
    cdef list t_tstop
    cdef list y_tstop
    cdef list yp_tstop

    cdef int order, nsteps
    cdef DTYPE_t maxstep, first_step
    cdef exclude_algvar_from_error, out
    cdef int compute_initcond
    cdef DTYPE_t compute_initcond_t0
    cdef long int mupper, mlower
    # ??? lband, uband, tcrit
    # ??? constraint_type, algebraic_var
    cdef bint initialized

    cdef void* _ida_mem
    cdef SUNContext sunctx
    cdef dict options
    cdef bint parallel_implementation
    cdef bint _old_api, _step_compute, _validate_flags
    cdef sunrealtype t, t0

    cdef IDA_data aux_data

    cdef int verbosity

    #cdef sunrealtype *y0, *yprime0

    # Functions
    cpdef _create_suncontext(self)
    cpdef _init_step(self, DTYPE_t t0,
                    np.ndarray[DTYPE_t, ndim=1] y0,
                    np.ndarray[DTYPE_t, ndim=1] yp0,
                    np.ndarray y_ic0_retn = ?,
                    np.ndarray yp_ic0_retn = ?)
    cpdef _reinit_IC(self, DTYPE_t t0,
                     np.ndarray[DTYPE_t, ndim=1] y0,
                     np.ndarray[DTYPE_t, ndim=1] yp0)
    cpdef _solve(self, np.ndarray[DTYPE_t, ndim=1] tspan,
                       np.ndarray[DTYPE_t, ndim=1] y0,
                       np.ndarray[DTYPE_t, ndim=1] yp0)
    cpdef _set_runtime_changeable_options(self, object options,
                                          bint supress_supported_check=*)



from .c_sundials cimport *
from libc.stdio cimport FILE

cdef extern from "cvodes/cvodes.h":
    # lmm
    enum: CV_ADAMS # 1
    enum: CV_BDF   # 2

    # itask
    enum: CV_NORMAL     # 1
    enum: CV_ONE_STEP   # 2

    # ism
    enum: CV_SIMULTANEOUS   # 1
    enum: CV_STAGGERED      # 2
    enum: CV_STAGGERED1     # 3

    # DQtype
    enum: CV_CENTERED       # 1
    enum: CV_FORWARD        # 2

    # interp
    enum: CV_HERMITE        # 1
    enum: CV_POLYNOMIAL     # 2

    # CVODE return flags
    enum: CV_SUCCESS           #    0
    enum: CV_TSTOP_RETURN      #    1
    enum: CV_ROOT_RETURN       #    2

    enum: CV_WARNING           #   99

    enum: CV_TOO_MUCH_WORK     #   -1
    enum: CV_TOO_MUCH_ACC      #   -2
    enum: CV_ERR_FAILURE       #   -3
    enum: CV_CONV_FAILURE      #   -4

    enum: CV_LINIT_FAIL        #   -5
    enum: CV_LSETUP_FAIL       #   -6
    enum: CV_LSOLVE_FAIL       #   -7
    enum: CV_RHSFUNC_FAIL      #   -8
    enum: CV_FIRST_RHSFUNC_ERR #   -9
    enum: CV_REPTD_RHSFUNC_ERR #   -10
    enum: CV_UNREC_RHSFUNC_ERR #   -11
    enum: CV_RTFUNC_FAIL       #   -12
    enum: CV_NLS_INIT_FAIL     #   -13
    enum: CV_NLS_SETUP_FAIL    #   -14
    enum: CV_CONSTR_FAIL       #   -15
    enum: CV_NLS_FAIL          #   -16

    enum: CV_MEM_FAIL          #   -20
    enum: CV_MEM_NULL          #   -21
    enum: CV_ILL_INPUT         #   -22
    enum: CV_NO_MALLOC         #   -23
    enum: CV_BAD_K             #   -24
    enum: CV_BAD_T             #   -25
    enum: CV_BAD_DKY           #   -26
    enum: CV_TOO_CLOSE         #   -27
    enum: CV_VECTOROP_ERR      #   -28
    
    enum: CV_NO_QUAD             # -30
    enum: CV_QRHSFUNC_FAIL       # -31
    enum: CV_FIRST_QRHSFUNC_ERR  # -32
    enum: CV_REPTD_QRHSFUNC_ERR  # -33
    enum: CV_UNREC_QRHSFUNC_ERR  # -34

    enum: CV_NO_SENS             # -40
    enum: CV_SRHSFUNC_FAIL       # -41
    enum: CV_FIRST_SRHSFUNC_ERR  # -42
    enum: CV_REPTD_SRHSFUNC_ERR  # -43
    enum: CV_UNREC_SRHSFUNC_ERR  # -44

    enum: CV_BAD_IS              # -45

    enum: CV_NO_QUADSENS         # -50
    enum: CV_QSRHSFUNC_FAIL      # -51
    enum: CV_FIRST_QSRHSFUNC_ERR # -52
    enum: CV_REPTD_QSRHSFUNC_ERR # -53
    enum: CV_UNREC_QSRHSFUNC_ERR # -54
    
    enum: CV_UNRECOGNIZED_ERR    # -99
    
    # adjoint return values 

    enum: CV_NO_ADJ             # -101
    enum: CV_NO_FWD             # -102
    enum: CV_NO_BCK             # -103
    enum: CV_BAD_TB0            # -104
    enum: CV_REIFWD_FAIL        # -105
    enum: CV_FWD_FAIL           # -106
    enum: CV_GETY_BADT          # -107

    ctypedef int (*CVRhsFn)(sunrealtype t, N_Vector y, N_Vector ydot, void *user_data) except? -1
    ctypedef int (*CVRootFn)(sunrealtype t, N_Vector y, sunrealtype *gout, void *user_data) except? -1
    ctypedef int (*CVEwtFn)(N_Vector y, N_Vector ewt, void *user_data)
    ctypedef void (*CVErrHandlerFn)(int error_code,
                               char *module, char *function,
                               char *msg, void *user_data)

    ctypedef int (*CVQuadRhsFn)(sunrealtype t, N_Vector y,
                                N_Vector yQdot, void *user_data)
    ctypedef int (*CVSensRhsFn)(int Ns, sunrealtype t,  N_Vector y, N_Vector ydot,
                                N_Vector *yS, N_Vector *ySdot,
                                void *user_data,
                                N_Vector tmp1, N_Vector tmp2)
    ctypedef int (*CVSensRhs1Fn)(int Ns, sunrealtype t, N_Vector y, N_Vector ydot,
                                 int iS, N_Vector yS, N_Vector ySdot,
                                 void *user_data,
                                 N_Vector tmp1, N_Vector tmp2)
    ctypedef int (*CVQuadSensRhsFn)(int Ns, sunrealtype t, N_Vector y, 
                                    N_Vector *yS, N_Vector yQdot, 
                                    N_Vector *yQSdot, void *user_data,
                                    N_Vector tmp, N_Vector tmpQ)
    ctypedef int (*CVRhsFnB)(sunrealtype t, N_Vector y, N_Vector yB,
                             N_Vector yBdot, void *user_dataB)
    ctypedef int (*CVRhsFnBS)(sunrealtype t, N_Vector y, N_Vector *yS,
                              N_Vector yB, N_Vector yBdot, void *user_dataB)

    ctypedef int (*CVQuadRhsFnB)(sunrealtype t, N_Vector y, N_Vector yB,
                                 N_Vector qBdot, void *user_dataB)
    ctypedef int (*CVQuadRhsFnBS)(sunrealtype t, N_Vector y, N_Vector *yS,
                                 N_Vector yB, N_Vector qBdot, void *user_dataB)

    # Exported Functions -- Forward Problems
    void *CVodeCreate(int lmm, SUNContext sunctx)

    int CVodeInit(void *cvode_mem, CVRhsFn f, sunrealtype t0, N_Vector y0)
    int CVodeReInit(void *cvode_mem, sunrealtype t0, N_Vector y0)

    int CVodeSStolerances(void *cvode_mem, sunrealtype reltol, sunrealtype abstol)
    int CVodeSVtolerances(void *cvode_mem, sunrealtype reltol, N_Vector abstol)
    int CVodeWFtolerances(void *cvode_mem, CVEwtFn efun)

    int CVodeSetErrHandlerFn(void *cvode_mem, CVErrHandlerFn ehfun, void *eh_data)
    int CVodeSetErrFile(void *cvode_mem, FILE *errfp)
    int CVodeSetUserData(void *cvode_mem, void *user_data)
    int CVodeSetMaxOrd(void *cvode_mem, int maxord)
    int CVodeSetMaxNumSteps(void *cvode_mem, long int mxsteps)
    int CVodeSetMaxHnilWarns(void *cvode_mem, int mxhnil)
    int CVodeSetStabLimDet(void *cvode_mem, sunbooleantype stldet)
    int CVodeSetInitStep(void *cvode_mem, sunrealtype hin)
    int CVodeSetMinStep(void *cvode_mem, sunrealtype hmin)
    int CVodeSetMaxStep(void *cvode_mem, sunrealtype hmax)
    int CVodeSetStopTime(void *cvode_mem, sunrealtype tstop)
    int CVodeSetMaxErrTestFails(void *cvode_mem, int maxnef)
    int CVodeSetMaxNonlinIters(void *cvode_mem, int maxcor)
    int CVodeSetMaxConvFails(void *cvode_mem, int maxncf)
    int CVodeSetNonlinConvCoef(void *cvode_mem, sunrealtype nlscoef)
    int CVodeSetConstraints(void *cvode_mem, N_Vector constraints)

    int CVodeSetNonlinearSolver(void *cvode_mem, SUNNonlinearSolver NLS)

    int CVodeRootInit(void *cvode_mem, int nrtfn, CVRootFn g)
    int CVodeSetRootDirection(void *cvode_mem, int *rootdir)
    int CVodeSetNoInactiveRootWarn(void *cvode_mem)

    int CVode(void *cvode_mem, sunrealtype tout, N_Vector yout,
                          sunrealtype *tret, int itask)
    int CVodeGetDky(void *cvode_mem, sunrealtype t, int k, N_Vector dky)

    int CVodeGetWorkSpace(void *cvode_mem, long int *lenrw, long int *leniw)
    int CVodeGetNumSteps(void *cvode_mem, long int *nsteps)
    int CVodeGetNumRhsEvals(void *cvode_mem, long int *nfevals)
    int CVodeGetNumLinSolvSetups(void *cvode_mem, long int *nlinsetups)
    int CVodeGetNumErrTestFails(void *cvode_mem, long int *netfails)
    int CVodeGetLastOrder(void *cvode_mem, int *qlast)
    int CVodeGetCurrentOrder(void *cvode_mem, int *qcur)
    int CVodeGetCurrentGamma(void *cvode_mem, sunrealtype *gamma)
    int CVodeGetNumStabLimOrderReds(void *cvode_mem, long int *nslred)
    int CVodeGetActualInitStep(void *cvode_mem, sunrealtype *hinused)
    int CVodeGetLastStep(void *cvode_mem, sunrealtype *hlast)
    int CVodeGetCurrentStep(void *cvode_mem, sunrealtype *hcur)
    int CVodeGetCurrentState(void *cvode_mem, N_Vector *y)
    int CVodeGetCurrentTime(void *cvode_mem, sunrealtype *tcur)
    int CVodeGetTolScaleFactor(void *cvode_mem, sunrealtype *tolsfac)
    int CVodeGetErrWeights(void *cvode_mem, N_Vector eweight)
    int CVodeGetEstLocalErrors(void *cvode_mem, N_Vector ele)
    int CVodeGetNumGEvals(void *cvode_mem, long int *ngevals)
    int CVodeGetRootInfo(void *cvode_mem, int *rootsfound)

    int CVodeGetIntegratorStats(void *cvode_mem, long int *nsteps,
                                            long int *nfevals, long int *nlinsetups,
                                            long int *netfails, int *qlast,
                                            int *qcur, sunrealtype *hinused, sunrealtype *hlast,
                                            sunrealtype *hcur, sunrealtype *tcur)

    int CVodeGetNumNonlinSolvIters(void *cvode_mem, long int *nniters)
    int CVodeGetNumNonlinSolvConvFails(void *cvode_mem, long int *nncfails)
    int CVodeGetNonlinSolvStats(void *cvode_mem, long int *nniters,
                                            long int *nncfails)
    char *CVodeGetReturnFlagName(long int flag)
    void CVodeFree(void **cvode_mem)

    # Exported Functions -- Quadrature

    int CVodeQuadInit(void *cvode_mem, CVQuadRhsFn fQ, N_Vector yQ0)
    int CVodeQuadReInit(void *cvode_mem, N_Vector yQ0)
    int CVodeQuadSStolerances(void *cvode_mem, sunrealtype reltolQ,
                              sunrealtype abstolQ)
    int CVodeQuadSVtolerances(void *cvode_mem, sunrealtype reltolQ,
                              N_Vector abstolQ)

    int CVodeSetQuadErrCon(void *cvode_mem, sunbooleantype errconQ)

    int CVodeGetQuad(void *cvode_mem, sunrealtype *tret, N_Vector yQout)
    int CVodeGetQuadDky(void *cvode_mem, sunrealtype t, int k, N_Vector dky)

    int CVodeGetQuadNumRhsEvals(void *cvode_mem, long int *nfQevals)
    int CVodeGetQuadNumErrTestFails(void *cvode_mem, long int *nQetfails)
    int CVodeGetQuadErrWeights(void *cvode_mem, N_Vector eQweight)
    int CVodeGetQuadStats(void *cvode_mem, long int *nfQevals, 
                          long int *nQetfails)

    void CVodeQuadFree(void *cvode_mem)

    # Exported Functions -- Sensitivities
    int CVodeSensInit(void *cvode_mem, int Ns, int ism,
                      CVSensRhsFn fS, N_Vector *yS0)
    int CVodeSensInit1(void *cvode_mem, int Ns, int ism,
                       CVSensRhs1Fn fS1, N_Vector *yS0)
    int CVodeSensReInit(void *cvode_mem, int ism, N_Vector *yS0)

    int CVodeSensSStolerances(void *cvode_mem, sunrealtype reltolS,
                              sunrealtype *abstolS)
    int CVodeSensSVtolerances(void *cvode_mem, sunrealtype reltolS,
                              N_Vector *abstolS)
    int CVodeSensEEtolerances(void *cvode_mem)

    int CVodeSetSensDQMethod(void *cvode_mem, int DQtype, sunrealtype DQrhomax)
    int CVodeSetSensErrCon(void *cvode_mem, sunbooleantype errconS)
    int CVodeSetSensMaxNonlinIters(void *cvode_mem, int maxcorS)
    int CVodeSetSensParams(void *cvode_mem, sunrealtype *p,
                           sunrealtype *pbar, int *plist)

    int CVodeSetNonlinearSolverSensSim(void *cvode_mem,
                                       SUNNonlinearSolver NLS)
    int CVodeSetNonlinearSolverSensStg(void *cvode_mem,
                                       SUNNonlinearSolver NLS)
    int CVodeSetNonlinearSolverSensStg1(void *cvode_mem,
                                        SUNNonlinearSolver NLS)

    int CVodeSensToggleOff(void *cvode_mem)

    int CVodeGetSens(void *cvode_mem, sunrealtype *tret, N_Vector *ySout)
    int CVodeGetSens1(void *cvode_mem, sunrealtype *tret, int iss,
                      N_Vector ySout) # rename is to iss, is protected in python
    int CVodeGetSensDky(void *cvode_mem, sunrealtype t, int k,
                        N_Vector *dkyA)
    int CVodeGetSensDky1(void *cvode_mem, sunrealtype t, int k, int iss,
                         N_Vector dky) # rename is to iss, is protected in python

    int CVodeGetSensNumRhsEvals(void *cvode_mem, long int *nfSevals)
    int CVodeGetNumRhsEvalsSens(void *cvode_mem, long int *nfevalsS)
    int CVodeGetSensNumErrTestFails(void *cvode_mem, long int *nSetfails)
    int CVodeGetSensNumLinSolvSetups(void *cvode_mem, long int *nlinsetupsS)
    int CVodeGetSensErrWeights(void *cvode_mem, N_Vector *eSweight)
    int CVodeGetSensStats(void *cvode_mem, long int *nfSevals,
                          long int *nfevalsS, long int *nSetfails,
                          long int *nlinsetupsS)
    int CVodeGetSensNumNonlinSolvIters(void *cvode_mem, long int *nSniters)
    int CVodeGetSensNumNonlinSolvConvFails(void *cvode_mem, long int *nSncfails)
    int CVodeGetStgrSensNumNonlinSolvIters(void *cvode_mem,
                                           long int *nSTGR1niters)
    int CVodeGetStgrSensNumNonlinSolvConvFails(void *cvode_mem,
                                               long int *nSTGR1ncfails)
    int CVodeGetSensNonlinSolvStats(void *cvode_mem, long int *nSniters,
                                    long int *nSncfails)

    void CVodeSensFree(void *cvode_mem)

    #Exported Functions -- Sensitivity dependent quadrature
    int CVodeQuadSensInit(void *cvode_mem, CVQuadSensRhsFn fQS,
                          N_Vector *yQS0)
    int CVodeQuadSensReInit(void *cvode_mem, N_Vector *yQS0)

    int CVodeQuadSensSStolerances(void *cvode_mem, sunrealtype reltolQS,
                                  sunrealtype *abstolQS)
    int CVodeQuadSensSVtolerances(void *cvode_mem, sunrealtype reltolQS,
                                  N_Vector *abstolQS)
    int CVodeQuadSensEEtolerances(void *cvode_mem)

    int CVodeSetQuadSensErrCon(void *cvode_mem, sunbooleantype errconQS)

    int CVodeGetQuadSens(void *cvode_mem, sunrealtype *tret, N_Vector *yQSout)
    int CVodeGetQuadSens1(void *cvode_mem, sunrealtype *tret, int iss,
                          N_Vector yQSout) # rename is to iss, is protected in python

    int CVodeGetQuadSensDky(void *cvode_mem, sunrealtype t, int k,
                            N_Vector *dkyQS_all)
    int CVodeGetQuadSensDky1(void *cvode_mem, sunrealtype t, int k,
                             int iss, N_Vector dkyQS) # rename is to iss, is protected in python

    int CVodeGetQuadSensNumRhsEvals(void *cvode_mem, long int *nfQSevals)
    int CVodeGetQuadSensNumErrTestFails(void *cvode_mem,
                                        long int *nQSetfails)
    int CVodeGetQuadSensErrWeights(void *cvode_mem, N_Vector *eQSweight)
    int CVodeGetQuadSensStats(void *cvode_mem, long int *nfQSevals,
                              long int *nQSetfails)

    void CVodeQuadSensFree(void *cvode_mem)

    # Exported Functions -- Backward Problems
    int CVodeAdjInit(void *cvode_mem, long int steps, int interp)
    int CVodeAdjReInit(void *cvode_mem)
    void CVodeAdjFree(void *cvode_mem)

    int CVodeCreateB(void *cvode_mem, int lmmB, int *which)
    int CVodeInitB(void *cvode_mem, int which, CVRhsFnB fB, sunrealtype tB0, 
                   N_Vector yB0)
    int CVodeInitBS(void *cvode_mem, int which, CVRhsFnBS fBs,
                    sunrealtype tB0, N_Vector yB0)
    int CVodeReInitB(void *cvode_mem, int which, sunrealtype tB0, N_Vector yB0)
    int CVodeSStolerancesB(void *cvode_mem, int which, sunrealtype reltolB,
                           sunrealtype abstolB)
    int CVodeSVtolerancesB(void *cvode_mem, int which,
                           sunrealtype reltolB, N_Vector abstolB)

    int CVodeQuadInitB(void *cvode_mem, int which,
                       CVQuadRhsFnB fQB, N_Vector yQB0)
    int CVodeQuadInitBS(void *cvode_mem, int which,
                        CVQuadRhsFnBS fQBs, N_Vector yQB0)
    int CVodeQuadReInitB(void *cvode_mem, int which, N_Vector yQB0)

    int CVodeQuadSStolerancesB(void *cvode_mem, int which,
                               sunrealtype reltolQB, sunrealtype abstolQB)
    int CVodeQuadSVtolerancesB(void *cvode_mem, int which,
                               sunrealtype reltolQB, N_Vector abstolQB)

    # Solver Function For Forward Problems 
    int CVodeF(void *cvode_mem, sunrealtype tout, N_Vector yout,
               sunrealtype *tret, int itask, int *ncheckPtr)

    # Solver Function For Backward Problems
    int CVodeB(void *cvode_mem, sunrealtype tBout, int itaskB)

    # Optional Input Functions For Adjoint Problems
    int CVodeSetAdjNoSensi(void *cvode_mem)
    
    int CVodeSetUserDataB(void *cvode_mem, int which, void *user_dataB)
    int CVodeSetMaxOrdB(void *cvode_mem, int which, int maxordB)
    int CVodeSetMaxNumStepsB(void *cvode_mem, int which, long int mxstepsB)
    int CVodeSetStabLimDetB(void *cvode_mem, int which, sunbooleantype stldetB)
    int CVodeSetInitStepB(void *cvode_mem, int which, sunrealtype hinB)
    int CVodeSetMinStepB(void *cvode_mem, int which, sunrealtype hminB)
    int CVodeSetMaxStepB(void *cvode_mem, int which, sunrealtype hmaxB)
    int CVodeSetConstraintsB(void *cvode_mem, int which, N_Vector constraintsB)
    int CVodeSetQuadErrConB(void *cvode_mem, int which, sunbooleantype errconQB)

    int CVodeSetNonlinearSolverB(void *cvode_mem, int which,
                                 SUNNonlinearSolver NLS)

    # Extraction And Dense Output Functions For Backward Problems
    int CVodeGetB(void *cvode_mem, int which, sunrealtype *tBret, N_Vector yB)
    int CVodeGetQuadB(void *cvode_mem, int which, sunrealtype *tBret, N_Vector qB)

    # Optional Output Functions For Backward Problems 
    void *CVodeGetAdjCVodeBmem(void *cvode_mem, int which)
    int CVodeGetAdjY(void *cvode_mem, sunrealtype t, N_Vector y)

    struct _CVadjCheckPointRec:
        void *my_addr
        void *next_addr
        sunrealtype t0
        sunrealtype t1
        long int nstep
        int order
        sunrealtype step

    ctypedef _CVadjCheckPointRec CVadjCheckPointRec

    int CVodeGetAdjCheckPointsInfo(void *cvode_mem, CVadjCheckPointRec *ckpnt)

    int CVodeGetAdjDataPointHermite(void *cvode_mem, int which,
                                    sunrealtype *t, N_Vector y, N_Vector yd)

    int CVodeGetAdjDataPointPolynomial(void *cvode_mem, int which,
                                       sunrealtype *t, int *order, N_Vector y)


    int CVodeGetAdjCurrentCheckPoint(void *cvode_mem, void **addr)

    
cdef extern from "cvodes/cvodes_ls.h":
    #CVDLS return values
    enum: CVLS_SUCCESS         #  0
    enum: CVLS_MEM_NULL        # -1
    enum: CVLS_LMEM_NULL       # -2
    enum: CVLS_ILL_INPUT       # -3
    enum: CVLS_MEM_FAIL        # -4
    enum: CVLS_PMEM_NULL        # -5
    enum: CVLS_JACFUNC_UNRECVR  # -6
    enum: CVLS_JACFUNC_RECVR    # -7
    enum: CVLS_SUNMAT_FAIL      # -8
    enum: CVLS_SUNLS_FAIL       # -9

    enum: CVLS_NO_ADJ           # -101
    enum: CVLS_LMEMB_NULL       # -102
    
    ctypedef int (*CVLsJacFn)(sunrealtype t, N_Vector y, N_Vector fy,
                              SUNMatrix Jac, void *user_data,
                              N_Vector tmp1, N_Vector tmp2, N_Vector tmp3) except? -1

    ctypedef int (*CVLsPrecSetupFn)(sunrealtype t, N_Vector y, N_Vector fy,
                                    sunbooleantype jok, sunbooleantype *jcurPtr,
                                    sunrealtype gamma, void *user_data) except? -1

    ctypedef int (*CVLsPrecSolveFn)(sunrealtype t, N_Vector y, N_Vector fy,
                                    N_Vector r, N_Vector z, sunrealtype gamma,
                                    sunrealtype delta, int lr, void *user_data) except? -1

    ctypedef int (*CVLsJacTimesSetupFn)(sunrealtype t, N_Vector y,
                                       N_Vector fy, void *user_data) except? -1

    ctypedef int (*CVLsJacTimesVecFn)(N_Vector v, N_Vector Jv, sunrealtype t,
                                      N_Vector y, N_Vector fy,
                                      void *user_data, N_Vector tmp) except? -1

    ctypedef int (*CVLsLinSysFn)(sunrealtype t, N_Vector y, N_Vector fy, 
                                 SUNMatrix A, sunbooleantype jok, 
                                 sunbooleantype *jcur, sunrealtype gamma,
                                 void *user_data, N_Vector tmp1, N_Vector tmp2,
                                 N_Vector tmp3)


    int CVodeSetLinearSolver(void *cvode_mem, SUNLinearSolver LS, SUNMatrix A)

    int CVodeSetJacFn(void *cvode_mem, CVLsJacFn jac)
    int CVodeSetMaxStepsBetweenJac(void *cvode_mem, long int msbj)
    int CVodeSetEpsLin(void *cvode_mem, sunrealtype eplifac)
    int CVodeSetPreconditioner(void *cvode_mem, CVLsPrecSetupFn pset,
                               CVLsPrecSolveFn psolve)
    int CVodeSetJacTimes(void *cvode_mem, CVLsJacTimesSetupFn jtsetup,
                         CVLsJacTimesVecFn jtimes)
    int CVodeSetLinSysFn(void *cvode_mem, CVLsLinSysFn linsys)

    int CVodeGetLinWorkSpace(void *cvode_mem, long int *lenrwLS,
                             long int *leniwLS)
    int CVodeGetNumJacEvals(void *cvode_mem, long int *njevals)
    int CVodeGetNumPrecEvals(void *cvode_mem, long int *npevals)
    int CVodeGetNumPrecSolves(void *cvode_mem, long int *npsolves)
    int CVodeGetNumLinIters(void *cvode_mem, long int *nliters)
    int CVodeGetNumLinConvFails(void *cvode_mem, long int *nlcfails)
    int CVodeGetNumJTSetupEvals(void *cvode_mem, long int *njtsetups)
    int CVodeGetNumJtimesEvals(void *cvode_mem, long int *njvevals)
    int CVodeGetNumLinRhsEvals(void *cvode_mem, long int *nfevalsLS)
    int CVodeGetLastLinFlag(void *cvode_mem, long int *flag)
    char *CVodeGetLinReturnFlagName(long int flag)

    #  Backward problems
    ctypedef int (*CVLsJacFnB)(sunrealtype t, N_Vector y, N_Vector yB,
                               N_Vector fyB, SUNMatrix JB,
                               void *user_dataB, N_Vector tmp1B,
                               N_Vector tmp2B, N_Vector tmp3B)

    ctypedef int (*CVLsJacFnBS)(sunrealtype t, N_Vector y, N_Vector *yS,
                                N_Vector yB, N_Vector fyB, SUNMatrix JB,
                                void *user_dataB, N_Vector tmp1B,
                                N_Vector tmp2B, N_Vector tmp3B)

    ctypedef int (*CVLsPrecSetupFnB)(sunrealtype t, N_Vector y, N_Vector yB,
                                     N_Vector fyB, sunbooleantype jokB,
                                     sunbooleantype *jcurPtrB,
                                     sunrealtype gammaB, void *user_dataB)

    ctypedef int (*CVLsPrecSetupFnBS)(sunrealtype t, N_Vector y,
                                      N_Vector *yS, N_Vector yB,
                                      N_Vector fyB, sunbooleantype jokB,
                                      sunbooleantype *jcurPtrB,
                                      sunrealtype gammaB, void *user_dataB)

    ctypedef int (*CVLsPrecSolveFnB)(sunrealtype t, N_Vector y, N_Vector yB,
                                     N_Vector fyB, N_Vector rB,
                                     N_Vector zB, sunrealtype gammaB,
                                     sunrealtype deltaB, int lrB,
                                     void *user_dataB)

    ctypedef int (*CVLsPrecSolveFnBS)(sunrealtype t, N_Vector y, N_Vector *yS,
                                      N_Vector yB, N_Vector fyB,
                                      N_Vector rB, N_Vector zB,
                                      sunrealtype gammaB, sunrealtype deltaB,
                                      int lrB, void *user_dataB)

    ctypedef int (*CVLsJacTimesSetupFnB)(sunrealtype t, N_Vector y, N_Vector yB,
                                         N_Vector fyB, void *jac_dataB)

    ctypedef int (*CVLsJacTimesSetupFnBS)(sunrealtype t, N_Vector y,
                                          N_Vector *yS, N_Vector yB,
                                          N_Vector fyB, void *jac_dataB)

    ctypedef int (*CVLsJacTimesVecFnB)(N_Vector vB, N_Vector JvB, sunrealtype t,
                                       N_Vector y, N_Vector yB, N_Vector fyB,
                                       void *jac_dataB, N_Vector tmpB)

    ctypedef int (*CVLsJacTimesVecFnBS)(N_Vector vB, N_Vector JvB,
                                        sunrealtype t, N_Vector y, N_Vector *yS,
                                        N_Vector yB, N_Vector fyB,
                                        void *jac_dataB, N_Vector tmpB)

    ctypedef int (*CVLsLinSysFnB)(sunrealtype t, N_Vector y, N_Vector yB, 
                                  N_Vector fyB, SUNMatrix AB, sunbooleantype jokB,
                                  sunbooleantype *jcurB, sunrealtype gammaB, 
                                  void *user_dataB, N_Vector tmp1B,
                                  N_Vector tmp2B, N_Vector tmp3B)

    ctypedef int (*CVLsLinSysFnBS)(sunrealtype t, N_Vector y, N_Vector* yS,
                                   N_Vector yB, N_Vector fyB, SUNMatrix AB,
                                   sunbooleantype jokB, sunbooleantype *jcurB,
                                   sunrealtype gammaB, void *user_dataB, 
                                   N_Vector tmp1B,
                                   N_Vector tmp2B, N_Vector tmp3B)

    int CVodeSetLinearSolverB(void *cvode_mem, int which, SUNLinearSolver LS,
                              SUNMatrix A)

    int CVodeSetJacFnB(void *cvode_mem, int which, CVLsJacFnB jacB)
    int CVodeSetJacFnBS(void *cvode_mem, int which, CVLsJacFnBS jacBS)

    int CVodeSetEpsLinB(void *cvode_mem, int which, sunrealtype eplifacB)

    int CVodeSetPreconditionerB(void *cvode_mem, int which, 
                                CVLsPrecSetupFnB psetB,
                                CVLsPrecSolveFnB psolveB)
    int CVodeSetPreconditionerBS(void *cvode_mem, int which,
                                 CVLsPrecSetupFnBS psetBS,
                                 CVLsPrecSolveFnBS psolveBS)

    int CVodeSetJacTimesB(void *cvode_mem, int which,
                          CVLsJacTimesSetupFnB jtsetupB,
                          CVLsJacTimesVecFnB jtimesB)
    int CVodeSetJacTimesBS(void *cvode_mem, int which,
                           CVLsJacTimesSetupFnBS jtsetupBS,
                           CVLsJacTimesVecFnBS jtimesBS)

    int CVodeSetLinSysFnB(void *cvode_mem, int which, CVLsLinSysFnB linsys)
    int CVodeSetLinSysFnBS(void *cvode_mem, int which, CVLsLinSysFnBS linsys)

cdef extern from "cvodes/cvodes_direct.h":
    ctypedef CVLsJacFn CVodeJacFn
    ctypedef CVLsJacFnB CVodeJacFnB
    ctypedef CVLsJacFnBS CVodeJacFnBS

    int CVodeSetJacFn(void *cvode_mem, CVodeJacFn jac)
    int CVodeGetWorkSpace(void *cvode_mem, long int *lenrwLS, long int *leniwLS)
    int CVodeGetNumJacEvals(void *cvode_mem, long int *njevals)
    int CVodeGetNumRhsEvals(void *cvode_mem, long int *nfevalsLS)
    int CVodeGetLastFlag(void *cvode_mem, long int *flag)
    char *CVodeGetReturnFlagName(long int flag)

    int CVodeSetJacFnB(void *cvode_mem, int which, CVodeJacFnB jacB)
    int CVodeSetJacFnBS(void *cvode_mem, int which, CVodeJacFnBS jacBS)

cdef extern from "cvodes/cvodes_bandpre.h":
    int CVBandPrecInit(void *cvode_mem, sunindextype N, sunindextype mu,
                       sunindextype ml);
    int CVBandPrecGetWorkSpace(void *cvode_mem, long int *lenrwLS, 
                               long int *leniwLS)
    int CVBandPrecGetNumRhsEvals(void *cvode_mem, long int *nfevalsBP)

    int CVBandPrecInitB(void *cvode_mem, int which,
                        sunindextype nB, sunindextype muB, sunindextype mlB)

cdef extern from "cvodes/cvodes_diag.h":
    # CVDIAG return values
    enum: CVDIAG_SUCCESS         #  0
    enum: CVDIAG_MEM_NULL        # -1
    enum: CVDIAG_LMEM_NULL       # -2
    enum: CVDIAG_ILL_INPUT       # -3
    enum: CVDIAG_MEM_FAIL        # -4
    # Additional last_flag values 
    enum: CVDIAG_INV_FAIL        # -5
    enum: CVDIAG_RHSFUNC_UNRECVR # -6
    enum: CVDIAG_RHSFUNC_RECVR   # -7

    enum: CVDIAG_NO_ADJ          # -101
    
    int CVDiag(void *cvode_mem)
    int CVDiagGetWorkSpace(void *cvode_mem, long int *lenrwLS, long int *leniwLS)
    int CVDiagGetNumRhsEvals(void *cvode_mem, long int *nfevalsLS)
    int CVDiagGetLastFlag(void *cvode_mem, long int *flag)
    char *CVDiagGetReturnFlagName(long int flag)

    int CVDiagB(void *cvode_mem, int which)

cdef extern from "cvodes/cvodes_bbdpre.h":
    ctypedef int (*CVLocalFn)(sunindextype Nlocal, sunrealtype t, N_Vector y,
                              N_Vector g, void *user_data)
    ctypedef int (*CVCommFn)(sunindextype Nlocal, sunrealtype t, N_Vector y,
                             void *user_data)

    int CVBBDPrecInit(void *cvode_mem, sunindextype Nlocal,
                      sunindextype mudq, sunindextype mldq,
                      sunindextype mukeep, sunindextype mlkeep,
                      sunrealtype dqrely, CVLocalFn gloc, CVCommFn cfn)
    int CVBBDPrecReInit(void *cvode_mem, sunindextype mudq, sunindextype mldq,
                        sunrealtype dqrely)
    int CVBBDPrecGetWorkSpace(void *cvode_mem, long int *lenrwBBDP, 
                              long int *leniwBBDP)
    int CVBBDPrecGetNumGfnEvals(void *cvode_mem, long int *ngevalsBBDP)

    ctypedef int (*CVLocalFnB)(sunindextype NlocalB, sunrealtype t,
                               N_Vector y, N_Vector yB, N_Vector gB, 
                               void *user_dataB)

    ctypedef int (*CVCommFnB)(sunindextype NlocalB, sunrealtype t,
                              N_Vector y, N_Vector yB, void *user_dataB)

    int CVBBDPrecInitB(void *cvode_mem, int which, sunindextype NlocalB,
                       sunindextype mudqB, sunindextype mldqB,
                       sunindextype mukeepB, sunindextype mlkeepB,
                       sunrealtype dqrelyB, CVLocalFnB glocB, CVCommFnB cfnB)

    int CVBBDPrecReInitB(void *cvode_mem, int which,
                         sunindextype mudqB, sunindextype mldqB,
                         sunrealtype dqrelyB)

cdef extern from "cvodes/cvodes_spils.h":

    ctypedef CVLsPrecSetupFn CVodePrecSetupFn
    ctypedef CVLsPrecSolveFn CVodePrecSolveFn
    ctypedef CVLsJacTimesSetupFn CVodeJacTimesSetupFn
    ctypedef CVLsJacTimesVecFn CVodeJacTimesVecFn
    
    ctypedef CVLsPrecSetupFnB CVodePrecSetupFnB;
    ctypedef CVLsPrecSetupFnBS CVodePrecSetupFnBS;
    ctypedef CVLsPrecSolveFnB CVodePrecSolveFnB;
    ctypedef CVLsPrecSolveFnBS CVodePrecSolveFnBS;
    ctypedef CVLsJacTimesSetupFnB CVodeJacTimesSetupFnB;
    ctypedef CVLsJacTimesSetupFnBS CVodeJacTimesSetupFnBS;
    ctypedef CVLsJacTimesVecFnB CVodeJacTimesVecFnB;
    ctypedef CVLsJacTimesVecFnBS CVodeJacTimesVecFnBS;

    int CVodeSetEpsLin(void *cvode_mem, sunrealtype eplifac)
    int CVodeSetPreconditioner(void *cvode_mem, CVodePrecSetupFn pset,
                                 CVodePrecSolveFn psolve)
    int CVodeSetJacTimes(void *cvode_mem, CVodeJacTimesSetupFn jtsetup,
                           CVodeJacTimesVecFn jtimes)

    int CVodeGetWorkSpace(void *cvode_mem, long int *lenrwLS, long int *leniwLS)
    int CVodeGetNumPrecEvals(void *cvode_mem, long int *npevals)
    int CVodeGetNumPrecSolves(void *cvode_mem, long int *npsolves)
    int CVodeGetNumLinIters(void *cvode_mem, long int *nliters)
    int CVodeGetNumConvFails(void *cvode_mem, long int *nlcfails)
    int CVodeGetNumJTSetupEvals(void *cvode_mem, long int *njtsetups)
    int CVodeGetNumJtimesEvals(void *cvode_mem, long int *njvevals)
    int CVodeGetNumRhsEvals(void *cvode_mem, long int *nfevalsLS)
    int CVodeGetLastFlag(void *cvode_mem, long int *flag)
    char *CVodeGetReturnFlagName(long int flag)

    int CVodeSetEpsLinB(void *cvode_mem, int which, sunrealtype eplifacB)
    int CVodeSetPreconditionerB(void *cvode_mem, int which,
                                  CVodePrecSetupFnB psetB,
                                  CVodePrecSolveFnB psolveB)
    int CVodeSetPreconditionerBS(void *cvode_mem, int which,
                                   CVodePrecSetupFnBS psetBS,
                                   CVodePrecSolveFnBS psolveBS)
    int CVodeSetJacTimesB(void *cvode_mem, int which,
                            CVodeJacTimesSetupFnB jtsetupB,
                            CVodeJacTimesVecFnB jtimesB)
    int CVodeSetJacTimesBS(void *cvode_mem, int which,
                             CVodeJacTimesSetupFnBS jtsetupBS,
                             CVodeJacTimesVecFnBS jtimesBS)



cimport numpy as np
from .c_sundials cimport N_Vector, sunrealtype
from .common_defs cimport DTYPE_t

from .ida cimport (IDA_RhsFunction, IDA_WrapRhsFunction, IDA_RootFunction,
                   IDA_WrapRootFunction, IDA_JacRhsFunction,
                   IDA_WrapJacRhsFunction, IDA_PrecSetupFunction,
                   IDA_WrapPrecSetupFunction, IDA_PrecSolveFunction,
                   IDA_WrapPrecSolveFunction, IDA_JacTimesVecFunction,
                   IDA_WrapJacTimesVecFunction, IDA_JacTimesSetupFunction,
                   IDA_WrapJacTimesSetupFunction, IDA_ContinuationFunction,
                   IDA_ErrHandler, IDA_WrapErrHandler, IDA_data,
                   IDA)


cdef class IDAS_data(IDA_data):
    cdef np.ndarray yS_tmp, ySdot_tmp

cdef class IDAS(IDA):
    cdef N_Vector aStol
    cdef IDAS_data aux_dataS
    
    cdef int Ns   #sensitivity parameter size



# cython: embedsignature=True

import numpy as np
cimport numpy as np
import inspect
from .c_sundials cimport (
    sunrealtype, sunindextype, N_Vector, SUNDlsMat, sunbooleantype,
    SUNMatrix, SUNMatGetID, SUNMATRIX_DENSE, SUNMATRIX_BAND, SUNMATRIX_SPARSE,
    SUNMATRIX_CUSTOM,
)
from .c_nvector_serial cimport (
    N_VGetLength_Serial as nv_length_s, # use function not macro
    NV_DATA_S as nv_data_s,
)
from .c_sunmatrix cimport (
    SUNDenseMatrix_Rows, SUNDenseMatrix_Columns, SUNDenseMatrix_Column,
    SUNBandMatrix_Columns, SUNBandMatrix_UpperBandwidth,
    SUNBandMatrix_LowerBandwidth, SUNBandMatrix_Column,
)

from libc.stdio cimport stderr

include "sundials_config.pxi"

precision = SUNDIALS_FLOAT_TYPE
if SUNDIALS_FLOAT_TYPE == "single":
    from numpy import single as DTYPE
elif SUNDIALS_FLOAT_TYPE == "double":
    from numpy import double as DTYPE
elif SUNDIALS_FLOAT_TYPE == "extended":
    from numpy import longdouble as DTYPE
else:
    # fall back to double
    from numpy import double as DTYPE

index_precision = SUNDIALS_INDEX_SIZE
if SUNDIALS_INDEX_SIZE == "32":
    from numpy import int32 as INDEX_TYPE
elif SUNDIALS_INDEX_SIZE == "64":
    from numpy import int64 as INDEX_TYPE
else:
    from numpy import int64 as INDEX_TYPE

has_lapack = SUNDIALS_BLAS_LAPACK

ctypedef sunrealtype *DlsMat_col
ctypedef sunrealtype *nv_content_data_s

# N_Vector content access functions
# These lack an equivalent we can use from cython
cdef inline sunrealtype get_nv_ith_s(nv_content_data_s vcd_s, int i):
    return vcd_s[i]

cdef inline void set_nv_ith_s(nv_content_data_s vcd_s, int i,
                          sunrealtype new_value):
    vcd_s[i] = new_value


# Dense matrix: access functions

cdef inline int get_dense_N(SUNDlsMat A):
    return A.N


cdef inline int get_dense_M(SUNDlsMat A):
    return A.M


cdef inline int get_band_mu(SUNDlsMat A):
    return A.mu


cdef inline int get_band_ml(SUNDlsMat A):
    return A.ml

cdef inline sunrealtype* get_dense_col(SUNDlsMat A, int j):
    return (A.cols)[j]


cdef inline void set_dense_col(SUNDlsMat A, int j, sunrealtype *data):
    (A.cols)[j] = data


cdef inline sunrealtype get_dense_element(SUNDlsMat A, int i, int j):
    return (A.cols)[j][i]


cdef inline void set_dense_element(SUNDlsMat A, int i, int j, sunrealtype aij):
    (A.cols)[j][i] = aij


# Band matrix access functions
cdef inline DlsMat_col get_band_col(SUNDlsMat A, int j):
    return ((A.cols)[j] + (A.s_mu))


cdef inline void set_band_col(SUNDlsMat A, int j, sunrealtype *data):
    ((A.cols)[j]) = data


cdef inline sunrealtype get_band_col_elem(DlsMat_col col_j, int i, int j):
    return col_j[(i)-(j)]


cdef inline void set_band_col_elem(DlsMat_col col_j, int i, int j, sunrealtype aij):
    col_j[(i)-(j)] = aij

cdef inline sunrealtype get_band_element(SUNDlsMat A, int i, int j):
    return ((A.cols)[j][(i)-(j)+(A.s_mu)])


cdef inline void set_band_element(SUNDlsMat A, int i, int j, sunrealtype aij):
    (A.cols)[j][(i)-(j)+(A.s_mu)] = aij


# Public functions
cdef int nv_s2ndarray(N_Vector v, np.ndarray[DTYPE_t, ndim=1] a) except? -1:
    """ copy a serial N_Vector v to a numpy array a """
    cdef sunindextype N, i
    N = nv_length_s(v)
    cdef nv_content_data_s v_data = nv_data_s(v)

    for i in range(N):
      a[i] = get_nv_ith_s(v_data, i)

cdef int ndarray2nv_s(N_Vector v, np.ndarray[DTYPE_t, ndim=1] a) except? -1:
    """ copy a numpy array a to a serial N_Vector v t"""
    cdef unsigned int N, i
    N = nv_length_s(v)
    cdef nv_content_data_s v_data = nv_data_s(v)

    for i in range(N):
      set_nv_ith_s(v_data, i, a[i])

cdef int SUNMatrix2ndarray(SUNMatrix m, np.ndarray a) except? -1:
    """ copy a SUNMatrix m to a numpy array a """
    cdef sunindextype N, M, i, j, ml, mu, stride
    cdef nv_content_data_s v_col

    if SUNMatGetID(m) == SUNMATRIX_DENSE:
        N = SUNDenseMatrix_Columns(m)
        M = SUNDenseMatrix_Rows(m)

        for j in range(N):
            v_col = SUNDenseMatrix_Column(m, j)
            for i in range(M):
                a[i,j] = get_nv_ith_s(v_col, i)

    elif SUNMatGetID(m) == SUNMATRIX_BAND:
        N = SUNBandMatrix_Columns(m)
        ml = SUNBandMatrix_LowerBandwidth(m)
        mu = SUNBandMatrix_UpperBandwidth(m)
        stride = ml + mu + 1
        for j in range(N):
            v_col = SUNBandMatrix_Column(m, j)
            for i in range(stride):
                a[i,j] = v_col[i - mu]

    else:
        raise NotImplementedError("SUNMatrix type not supported")

cdef int ndarray2SUNMatrix(SUNMatrix m, np.ndarray a) except? -1:
    """ copy a numpy array a to a SUNMatrix m"""
    cdef sunindextype N, M, i, j, ml, mu, stride
    cdef nv_content_data_s v_col

    if SUNMatGetID(m) == SUNMATRIX_DENSE:
        N = SUNDenseMatrix_Columns(m)
        M = SUNDenseMatrix_Rows(m)

        for j in range(N):
            v_col = SUNDenseMatrix_Column(m, j)
            for i in range(M):
                v_col[i] = a[i,j]

    elif SUNMatGetID(m) == SUNMATRIX_BAND:
        N = SUNBandMatrix_Columns(m)
        ml = SUNBandMatrix_LowerBandwidth(m)
        mu = SUNBandMatrix_UpperBandwidth(m)
        stride = ml + mu + 1
        for j in range(N):
            v_col = SUNBandMatrix_Column(m, j)
            for i in range(stride):
                v_col[i - mu] = a[i,j]

    else:
        raise NotImplementedError("SUNMatrix type not supported")

cdef ensure_numpy_float_array(object value):
    try:
        if (type(value) == float or type(value) == int
            or type(value) == np.float
            or type(value) == np.float32
            or type(value) == np.float64
            or type(value) == np.float128
            or type(value) == np.int
            or type(value) == np.int8
            or type(value) == np.int16
            or type(value) == np.int32
            or type(value) == np.int64):

            return np.array([value, ], DTYPE)
        else:
            return np.asarray(value, DTYPE)
    except:
        raise ValueError('ensure_numpy_float_array: value not a number or '
                         'sequence of numbers: %s' % value)



# cython: embedsignature=True
from cpython.exc cimport PyErr_CheckSignals
from collections import namedtuple
from enum import IntEnum
import inspect
from warnings import warn

include "sundials_config.pxi"

import numpy as np
cimport numpy as np

from . import (
    CVODESolveFailed, CVODESolveFoundRoot, CVODESolveReachedTSTOP,
    _get_num_args,
)

from .c_sundials cimport sunrealtype, N_Vector
from .c_nvector_serial cimport *
from .c_sunmatrix cimport *
from .c_sunlinsol cimport *
from .c_sunnonlinsol cimport *

from .cvode cimport (CV_RhsFunction, CV_WrapRhsFunction, CV_RootFunction,
                       CV_WrapRootFunction, CV_JacRhsFunction,
                       CV_WrapJacRhsFunction, CV_PrecSetupFunction,
                       CV_WrapPrecSetupFunction,
                       CV_PrecSolveFunction, CV_WrapPrecSolveFunction,
                       CV_JacTimesVecFunction, CV_WrapJacTimesVecFunction,
                       CV_JacTimesSetupFunction, CV_WrapJacTimesSetupFunction,
                       CV_ContinuationFunction, CV_ErrHandler, 
                       CV_WrapErrHandler, CV_data, CVODE)
from .c_cvodes cimport *
from .common_defs cimport (
    nv_s2ndarray, ndarray2nv_s, ndarray2SUNMatrix, DTYPE_t, INDEX_TYPE_t,
)
from .common_defs import DTYPE, INDEX_TYPE
# this is needed because we want DTYPE and INDEX_TYPE to be
# accessible from python (not only in cython)


# TODO: parallel implementation: N_VectorParallel
# TODO: linsolvers: check the output value for errors
# TODO: optimize code for compiler

SolverReturn = namedtuple(
    "SolverReturn", [
        "flag", "values", "errors", "roots",
        "tstop", "message"
    ]
)

SolverVariables = namedtuple("SolverVariables", ["t", "y"])

class StatusEnum(IntEnum):
    SUCCESS           = CV_SUCCESS              # 0
    TSTOP_RETURN      = CV_TSTOP_RETURN         # 1
    ROOT_RETURN       = CV_ROOT_RETURN          # 2
    WARNING           = CV_WARNING              # 99
    TOO_MUCH_WORK     = CV_TOO_MUCH_WORK        # -1
    TOO_MUCH_ACC      = CV_TOO_MUCH_ACC         # -2
    ERR_FAILURE       = CV_ERR_FAILURE          # -3
    CONV_FAILURE      = CV_CONV_FAILURE         # -4
    LINIT_FAIL        = CV_LINIT_FAIL           # -5
    LSETUP_FAIL       = CV_LSETUP_FAIL          # -6
    LSOLVE_FAIL       = CV_LSOLVE_FAIL          # -7
    RHSFUNC_FAIL      = CV_RHSFUNC_FAIL         # -8
    FIRST_RHSFUNC_ERR = CV_FIRST_RHSFUNC_ERR    # -9
    REPTD_RHSFUNC_ERR = CV_REPTD_RHSFUNC_ERR    # -10
    UNREC_RHSFUNC_ERR = CV_UNREC_RHSFUNC_ERR    # -11
    RTFUNC_FAIL       = CV_RTFUNC_FAIL          # -12
    NLS_INIT_FAIL     = CV_NLS_INIT_FAIL        # -13
    NLS_SETUP_FAIL    = CV_NLS_SETUP_FAIL       # -14
    CONSTR_FAIL       = CV_CONSTR_FAIL          # -15
    NLS_FAIL          = CV_NLS_FAIL             # -16
    MEM_FAIL          = CV_MEM_FAIL             # -20
    MEM_NULL          = CV_MEM_NULL             # -21
    ILL_INPUT         = CV_ILL_INPUT            # -22
    NO_MALLOC         = CV_NO_MALLOC            # -23
    BAD_K             = CV_BAD_K                # -24
    BAD_T             = CV_BAD_T                # -25
    BAD_DKY           = CV_BAD_DKY              # -26
    TOO_CLOSE         = CV_TOO_CLOSE            # -27
    VECTOROP_ERR      = CV_VECTOROP_ERR         # -28
    NO_QUAD           = CV_NO_QUAD              # -30
    QRHSFUNC_FAIL     = CV_QRHSFUNC_FAIL        # -31
    FIRST_QRHSFUNC_ERR= CV_FIRST_QRHSFUNC_ERR   # -32
    REPTD_QRHSFUNC_ERR= CV_REPTD_QRHSFUNC_ERR   # -33
    UNREC_QRHSFUNC_ERR= CV_UNREC_QRHSFUNC_ERR   # -34
    NO_SENS           = CV_NO_SENS              # -40
    SRHSFUNC_FAIL     = CV_SRHSFUNC_FAIL        # -41
    FIRST_SRHSFUNC_ERR= CV_FIRST_SRHSFUNC_ERR   # -42
    REPTD_SRHSFUNC_ERR= CV_REPTD_SRHSFUNC_ERR   # -43
    UNREC_SRHSFUNC_ERR= CV_UNREC_SRHSFUNC_ERR   # -44
    BAD_IS            = CV_BAD_IS               # -45
    NO_QUADSENS       = CV_NO_QUADSENS          # -50
    QSRHSFUNC_FAIL    = CV_QSRHSFUNC_FAIL       # -51
    FIRST_QSRHSFUNC_ERR=CV_FIRST_QSRHSFUNC_ERR  # -52
    REPTD_QSRHSFUNC_ERR=CV_REPTD_QSRHSFUNC_ERR  # -53
    UNREC_QSRHSFUNC_ERR=CV_UNREC_QSRHSFUNC_ERR  # -54
    UNRECOGNIZED_ERR  = CV_UNRECOGNIZED_ERR     # -99

STATUS_MESSAGE = {
    StatusEnum.SUCCESS: "Successful function return.",
    StatusEnum.TSTOP_RETURN: "Reached specified stopping point",
    StatusEnum.ROOT_RETURN: "Found one or more roots",
    StatusEnum.WARNING: "Succeeded but something unusual happened",
    StatusEnum.TOO_MUCH_WORK: "Could not reach endpoint",
    StatusEnum.TOO_MUCH_ACC: "Could not satisfy accuracy",
    StatusEnum.ERR_FAILURE: "Error test failures occurred too many times during one internal time step or minimum step size was reached.",
    StatusEnum.CONV_FAILURE: "Convergence test failures occurred too many times during one internal time step or minimum step size was reached.",
    StatusEnum.LINIT_FAIL: "The linear solver’s initialization function failed.",
    StatusEnum.LSETUP_FAIL: "The linear solver’s setup function failed in an unrecoverable manner.",
    StatusEnum.LSOLVE_FAIL: "The linear solver’s solve function failed in an unrecoverable manner.",
    StatusEnum.RHSFUNC_FAIL: "The right-hand side function failed in an unrecoverable manner.",
    StatusEnum.FIRST_RHSFUNC_ERR: "The right-hand side function failed at the first call.",
    StatusEnum.REPTD_RHSFUNC_ERR: "The right-hand side function had repeated recoverable errors.",
    StatusEnum.UNREC_RHSFUNC_ERR: "The right-hand side function had a recoverable error, but no recovery is possible.",
    StatusEnum.RTFUNC_FAIL: "The rootfinding function failed in an unrecoverable manner.",
    StatusEnum.NLS_INIT_FAIL: "The nonlinear solver's init routine failed.",
    StatusEnum.NLS_SETUP_FAIL: "The nonlinear solver setup failed unrecoverably.",
    StatusEnum.CONSTR_FAIL: "The inequality constraints could not be met",
    StatusEnum.NLS_FAIL: "The nonlinear solver failed in an unrecoverable manner",
    StatusEnum.MEM_FAIL: "A memory allocation failed.",
    StatusEnum.MEM_NULL: "The cvode_mem argument was NULL.",
    StatusEnum.ILL_INPUT: "One of the function inputs is illegal.",
    StatusEnum.NO_MALLOC: "The cvode memory block was not allocated by a call to CVodeMalloc.",
    StatusEnum.BAD_K: "The derivative order k is larger than the order used.",
    StatusEnum.BAD_T: "The time t is outside the last step taken.",
    StatusEnum.BAD_DKY: "The output derivative vector is NULL.",
    StatusEnum.TOO_CLOSE: "The output and initial times are too close to each other.",
    StatusEnum.VECTOROP_ERR: "Vector operation error",
    StatusEnum.NO_QUAD: "Quadrature integration not activated.",
    StatusEnum.QRHSFUNC_FAIL: "The quadrature right-hand side routine failed in an unrecoverable manner.",
    StatusEnum.FIRST_QRHSFUNC_ERR: "The quadrature right-hand side routine failed at the first call.",
    StatusEnum.REPTD_QRHSFUNC_ERR: "Repeated recoverable quadrature right-hand side function errors.",
    StatusEnum.UNREC_QRHSFUNC_ERR: "The quadrature right-hand side failed in a recoverable manner, but no recovery is possible.",
    StatusEnum.NO_SENS: "Forward sensitivity analysis not activated.",
    StatusEnum.SRHSFUNC_FAIL: "The quadrature sensitivity right-hand side routine failed in an unrecoverable manner.",
    StatusEnum.FIRST_SRHSFUNC_ERR: "The sensitivity right-hand side routine failed at the first call.",
    StatusEnum.REPTD_SRHSFUNC_ERR: "Repeated recoverable sensitivity right-hand side function errors.",
    StatusEnum.UNREC_SRHSFUNC_ERR: "The sensitivity right-hand side failed in a recoverable manner, but no recovery is possible.",
    StatusEnum.BAD_IS: "Illegal value for is parameter.",
    StatusEnum.NO_QUADSENS: "Forward sensitivity analysis for quadrature variables not activated.",
    StatusEnum.QSRHSFUNC_FAIL: "The quadrature sensitivity right-hand side routine failed in an unrecoverable manner.",
    StatusEnum.FIRST_QSRHSFUNC_ERR: "The quadrature sensitivity right-hand side routine failed at the first call.",
    StatusEnum.REPTD_QSRHSFUNC_ERR: "Repeated recoverable quadrature sensitivity right-hand side function errors.",
    StatusEnum.UNREC_QSRHSFUNC_ERR: "The quadrature sensitivity right-hand side failed in a recoverable manner, but no recovery is possible.",
    StatusEnum.UNRECOGNIZED_ERR: "Unrecognized Error",
}

WARNING_STR = "Solver succeeded with flag {} and finished at {} with values {}"


# Auxiliary data carrying runtime vales for the CVODE solver
cdef class CVS_data(CV_data):
    def __cinit__(self, N, Ns=0):
        super(CVS_data, self).__init__(N)
        
        self.yS_tmp = np.empty(Ns, DTYPE)
        self.ySdot_tmp = np.empty(Ns, DTYPE)

cdef class CVODES(CVODE):

    def __cinit__(self, Rfn, **options):
        """
        Initialize the CVODE Solver and it's default values

        Input:
            Rfn     - right-hand-side function
            options - additional options for initialization, for the list
                      of supported options and their values see set_options()

        """
        super(CVODES, self).__init__(Rfn, **options)
        
        self.Ns       = -1

    def set_options(self, **options):
        """
        Reads the options list and assigns values for the solver.

        All options list:
            'verbosity':
                Values: 0,1,2,...
                Description:
                    Set the level of verbosity. The higher number user, the
                    more verbose the output will be. Default is 1.
                    Deprecated, does nothing
                Note: Changeable at runtime.
            'implementation':
                Values: 'serial' (= default), 'parallel'
                Description:
                    Using serial or parallel implementation of the solver.
                    #TODO: curently only serial implementation is working
            'lmm_type'  - linear multistep method
                Values: 'ADAMS', 'BDF' (= default)
                Description:
                    Recommended combination:

                    Problem  | 'lmm_type' | nonlinearsolver module
                    ------------------------------------
                    nonstiff | 'ADAMS'    | 'fixedpoint' (SUNNonlinearSolver_FixedPoint )
                    stiff    | 'BDF'      | 'newton' (SUNNonlinearSolver_Newton)

                    See also 'iter_type'.
            'nonlinsolver' - nonlinear solver iteration
                Values: 'newton' (= default), 'fixedpoint'
                Description:
                    See 'lmm_type'.

            'rfn':
                Values: function of class CV_RhsFunction or a python function
                        with signature (t, y, yp) or (t, y, yp, userdata)
                Description:
                    Defines the right-hand-side function (which has to be
                    a subclass of CV_RhsFunction class, or a normal python
                    function with signature (t, y, yp) or (t, y, yp, userdata)).
                    This function takes as input arguments current time t,
                    current value of y, and yp must be used as output numpy
                    array of returned values of \dot{y} at t.
                    Optional userdata. Return value is 0 if successfull.
                    This option is mandatory.
            'rootfn':
                Values: function of class CV_RootFunction or a python function
                    with signature (t, y, g, user_data)
                Description:
                    Defines a function that fills a vector 'g' with values
                    (g is a vector of size k). Vector 'g represents conditions
                    which when satisfied, the solver stops. Namely, the solver
                    stops when condition g[i] is zero for some i.
                    For 'user_data' argument see 'user_data' option
                    above.
                Note: Changeable at runtime.
            'nr_rootfns':
                Value: integer
                Description:
                    The length of the array returned by 'rootfn' (see above),
                    i.e. number of conditions for which we search the value 0.
                Note: Changeable at runtime.
            'jacfn':
                Values: function of class CV_JacRhsFunction
                Description:
                    Defines the jacobian function and has to be a subclass
                    of CV_JacRhsFunction class or python function. This function
                    takes as input arguments current time t, current value of y,
                    current value of f(t,y), and 
                    a 2D numpy array of returned jacobian and optional userdata.
                    Return value is 0 if successfull.
                    Jacobian is only used for dense or lapackdense linear solver
                TODO: cvode supports Jacobian for band also, this is not supported.
            'rtol':
                Values: float,  1e-6 = default
                Description:
                    Relative tolerancy.
                Note: Changeable at runtime.
            'atol':
                Values: float or numpy array of floats,  1e-12 = default
                Description:
                    Absolute tolerancy
                Note: Changeable at runtime.
            'order':
                Values: 1, 2, 3, 4, 5 (= default)
                Description:
                    Specifies the maximum order of the linear multistep method.
            'max_steps':
                Values: positive integer, 0 = default (uses value of 500)
                Description:
                    Maximum number of (internally defined) steps allowed
                    during one call to the solver.
            'max_step_size':
                Values: non-negative float, 0.0 = default
                Description:
                    Restricts the maximal (absolute) internal step value taken
                    by the solver. The value of 0.0 uses
                    no restriction on the maximal step size.
            'min_step_size':
                Values: non-negative float, 0.0 = default
                Description:
                    Restricts the minimal (absolute) internal step value taken
                    by the solver. The value of 0.0 uses
                    no restriction.
            'first_step_size':
                Values: float , 0.0 = default
                Description:
                    Sets the first step size. DAE solver can suffer on the
                    first step, so set this to circumvent this. The value
                    of 0.0 uses the solver's internal default value.
            'linsolver':
                Values: 'dense' (= default), 'lapackdense', 'band',
                        'lapackband', 'spgmr', 'spbcgs', 'sptfqmr'
                Description:
                    Specifies used linear solver.
                    Limitations: Linear solvers for dense and band matrices
                                 can be used only for serial implementation.
                                 For parallel implementation use_relaxation
                                 use lapackdense or lapackband respectively.
                    TODO: to add new solvers: pcg, spfgmr, superlumt, klu
                    
            'nonlinsolver':
                Values: 'newton' (= default), 'fixedpoint'
                Description:
                    Specifies the used nonlinear solver.
                    
            'lband', 'uband':
                Values: non-negative integer, 0 = default
                Description:
                    Specifies the width of the lower band (below the main
                    diagonal) and/or upper diagonal (above the main diagonal).
                    So the total band width is lband + 1 + uband, where the 1
                    stands for the main diagonal
                    (specially, lband = uband = 0 [that is the default]
                    denotes the band width = 1, i.e. the main diagonal only).
                    Used only if 'linsolver' is band.
            'maxl':
                Values: 0 (= default), 1, 2, 3, 4, 5
                Description:
                    Dimension of the number of used Krylov subspaces
                    (used only by 'spgmr', 'spbcgs', 'sptfqmr' linsolvers)
            'tstop':
                Values: float, 0.0 = default
                Description:
                    Maximum time until which we perform the computations.
                    Default is 0.0. Once the value is set 0.0, it cannot be
                    disabled (but it will be automatically disabled when tstop
                    is reached and has to be reset again in next run).
                Note: Changeable at runtime.
            'one_step_compute':
                Values: False (default) or True
                Description:
                    Only influences the step() method.
                    Solver computes normally up to the required output time.
                    If value is True, solver insteads does one internal step
                    of the solver in the direction of the output time, as seen
                    from the current time of the solution.
            'user_data':
                Values: python object, None = default
                Description:
                    Additional data that are supplied to each call of the
                    residual function 'Rfn' (see below) and Jacobi function
                    'jacfn' (if specified one).
            'precond_type':
                default = None
            'prec_setupfn':
                Values: function of class CV_PrecSetupFunction
                Description:
                    Defines a function that setups the preconditioner on change
                    of the Jacobian. This function takes as input arguments
                    current time t, current value of y, flag jok that indicates
                    whether Jacobian-related data has to be updated, flag jcurPtr
                    that should be set to True (jcurPtr.value = True) if Jacobian
                    data was recomputed, parameter gamma and optional userdata.
            'prec_solvefn':
                Values: function of class CV_PrecSolveFunction
                Description:
                    Defines a function that solves the preconditioning problem
                    P*z = r where P may be a left or right preconditioner
                    matrix. This function takes as input arguments current time
                    t, current value of y, right-hand side r, result vector z,
                    parameters gamma and delta, input flag lr that determines
                    the flavour of the preconditioner (left = 1, right = 2) and
                    optional userdata.
            'jac_times_vecfn':
                Values: function of class CV_JacTimesVecFunction
                Description:
                    Defines a function that solves the product of vector v
                    with an (approximate) Jacobian of the system J.
                    This function takes as input arguments the vector v,
                    result vector Jv, current time t, current value of y,
                    and optional userdata.
            'jac_times_setupfn':
                Values: function of class CV_JacTimesSetupFunction
                Description:
                    Optional. Default is to internal finite difference with no
                    extra setup.
            'bdf_stability_detection':
                default = False, only used if lmm_type == 'bdf
            'max_conv_fails':
                default = 0,
            'max_nonlin_iters':
                default = 0,
            'nonlin_conv_coef':
                default = 0,
            'err_handler':
                Values: function of class CV_ErrHandler, default = None
                Description:
                    Defines a function which controls output from the CVODE
                    solver
            'err_user_data':
                Description:
                    User data used by 'err_handler', defaults to 'user_data'
            'old_api':
                Values: True (default), False
                Description:
                    Forces use of old api (tuple of 5, with last two items
                    overloaded) if True or new api (namedtuple) if False.
                    Other options may require new api, hence using this should
                    be avoided if possible.
            'onroot':
                Description:
                    If the solver returns ROOT, call this function. If it
                    returns 0, continue solving, otherwise stop.
                    If not given, solver stops after a ROOT
            'ontstop':
                Description:
                    If the solver returns TSTOP, call this function. If it
                    returns 0, continue solving, otherwise stop.
                    If not given, solver stops after a TSTOP
            'validate_flags':
                Values: True, False (=default)
                Description:
                    Controls whether to validate flags as a result of calling
                    `solve`. See the `validate_flags` function for how this
                    affects `solve`.
        """

        # Update values of all supplied options
        super(CVODES, self).set_options(**options)

    cpdef _set_runtime_changeable_options(self, object options,
                                          bint supress_supported_check=False):
        """
          (Re)Set options that can change after CV_MEM object was allocated -
          mostly useful for values that need to be changed at run-time.
        """
        
        super(CVODES, self)._set_runtime_changeable_options(options,
                                              supress_supported_check)

    def init_step(self, DTYPE_t t0, object y0):
        """
        Initialize the solver and all the internal variables. This assumes
        the call to 'set_options()' to be done and hence all the information
        for correct solver initialization to be available.
        An error is raised for almost all errors.

        Return values:
         if old_api:
            flag  - boolean status of the computation (successful or error occured)
            t_out - initial time

         if old_api False (cvode solver):
            A named tuple, with entries:
                flag   = An integer flag (StatusEnumXXX)
                values = Named tuple with entries t and y and ydot. y will
                            correspond to y_retn value and ydot to yp_retn!
                errors = Named tuple with entries t_err and y_err
                roots  = Named tuple with entries t_roots and y_roots
                tstop  = Named tuple with entries t_stop and y_tstop
                message= String with message in case of an error

        Note: some options can be re-set also at runtime. See 'reinit_IC()'
        """
        
        soln = super(CVODES, self).init_step(t0, y0)
        return soln

    def solve(self, object tspan, object y0):
        """
        Runs the solver.

        Input:
            tspan - an list/array of times at which the computed value will be
                    returned. Must contain the start time as first entry..
            y0    - list/numpy array of initial values

        Return values:
         if old_api
            flag   - indicating return status of the solver
            t      - numpy array of times at which the computations were successful
            y      - numpy array of values corresponding to times t (values of y[i, :] ~ t[i])
            t_err  - float or None - if recoverable error occured (for example reached maximum
                     number of allowed iterations), this is the time at which it happened
            y_err  - numpy array of values corresponding to time t_err
         if old_api False (cvode solver):
            A named tuple, with entries:
                flag   = An integer flag (StatusEnum)
                values = Named tuple with entries t and y
                errors = Named tuple with entries t and y
                roots  = Named tuple with entries t and y
                tstop  = Named tuple with entries t and y
                message= String with message in case of an error
        """
        soln = super(CVODES, self).solve(tspan, y0)
        return soln

    def step(self, DTYPE_t t, np.ndarray[DTYPE_t, ndim=1] y_retn = None):
        """
        Method for calling successive next step of the CVODE solver to allow
        more precise control over the solver. The 'init_step' method has to
        be called before the 'step' method.

        Input:
            t - if t>0.0 then integration is performed until this time
                         and results at this time are returned in y_retn
              - if t<0.0 only one internal step is perfomed towards time abs(t)
                         and results after this one time step are returned
            y_retn - numpy vector (ndim = 1) in which the computed
                     value will be stored  (needs to be preallocated)
        Return values:
            flag  - status of the computation (successful or error occured)
            t_out - time, where the solver stopped
                    (when no error occured, t_out == t)
        """
        soln = super(CVODES, self).step(t, y_retn)
        return soln

    def get_info(self):
        """
        Returns a dict with information about the solver, like number
        of calls to the user's right-hand side function.

        """
        info = super(CVODES, self).get_info()

        return info



import io
import os
from os.path import join
import sys
import textwrap

from distutils.log import info
from setuptools import Extension
from distutils.command.build_ext import build_ext as _build_ext

PKGCONFIG_CVODE = 'sundials-cvode-serial'
PKGCONFIG_IDA = 'sundials-ida-serial'
PKGCONFIG_CVODES = 'sundials-cvodes-serial'
PKGCONFIG_IDAS = 'sundials-idas-serial'


def write_pxi(filename, definitions):
    """
    Write a cython include file (.pxi), `filename`, with the definitions in the
    `definitions` mapping.
    """
    with io.open(filename, mode='w', encoding='utf-8') as pxi_file:
        for name, val in definitions.items():
            pxi_file.write(u"DEF {name} = {val}\n".format(name=name, val=val))
    return filename


def check_macro_true(cmd, symbol, headers=None, include_dirs=None):
    """
    Copied from numpy.distutils.command.config:config.check_macro_true, checks
    if macro is true
    """
    cmd._check_compiler()
    body = textwrap.dedent("""
        int main(void)
        {
        #if %s
        #else
        #error false or undefined macro
        #endif
            ;
            return 0;
        }""") % (symbol,)

    return cmd.try_compile(body, headers, include_dirs)


def check_macro_def(cmd, symbol, headers=None, include_dirs=None):
    """
    Based on numpy.distutils.command.config:config.check_macro_true, checks if
    macro is defined or not
    """
    cmd._check_compiler()
    body = textwrap.dedent("""
        int main(void)
        {
        #ifdef %s
        #else
        #error undefined macro
        #endif
            ;
            return 0;
        }""") % (symbol,)

    return cmd.try_compile(body, headers, include_dirs)


def get_sundials_config_pxi(include_dirs, dist):
    """
    Create pxi file containing some of sundials build config

    Don't ask why this is a function, something crazy about
    distutils/numpy not setting _setup_distribution at the right time or
    something...
    """
    SUNDIALS_CONFIG_H = "sundials/sundials_config.h"
    BASE_PATH = 'src/scikits_odes_sundials'

    config_cmd = dist.get_command_obj("config")

    # Get float type
    if check_macro_true(
        config_cmd,
        "SUNDIALS_DOUBLE_PRECISION", headers=[SUNDIALS_CONFIG_H],
        include_dirs=include_dirs
    ):
        SUNDIALS_FLOAT_TYPE = '"double"'
        info("Found sundials built with double precision.")
    elif check_macro_true(
        config_cmd,
        "SUNDIALS_SINGLE_PRECISION", headers=[SUNDIALS_CONFIG_H],
        include_dirs=include_dirs
    ):
        SUNDIALS_FLOAT_TYPE = '"single"'
        info("Found sundials built with single precision.")
    elif check_macro_true(
        config_cmd,
        "SUNDIALS_EXTENDED_PRECISION", headers=[SUNDIALS_CONFIG_H],
        include_dirs=include_dirs
    ):
        SUNDIALS_FLOAT_TYPE = '"extended"'
        info("Found sundials built with extended precision.")
    else:
        # fall back to double
        SUNDIALS_FLOAT_TYPE = '"double"'
        info("Failed to find sundials precision, falling back to double...")

    # Get index (int) type
    if check_macro_true(
        config_cmd,
        "SUNDIALS_INT32_T", headers=[SUNDIALS_CONFIG_H],
        include_dirs=include_dirs
    ):
        SUNDIALS_INDEX_SIZE = '"int32"'
        info("Found sundials built with int32.")
    elif check_macro_true(
        config_cmd,
        "SUNDIALS_INT64_T", headers=[SUNDIALS_CONFIG_H],
        include_dirs=include_dirs
    ):
        SUNDIALS_INDEX_SIZE = '"64"'
        info("Found sundials built with int64.")
    else:
        # fall back to int64
        SUNDIALS_INDEX_SIZE = '"64"'
        info("Failed to find sundials index type, falling back to int64...")

    # Check for blas/lapack
    if check_macro_def(
        config_cmd,
        "SUNDIALS_BLAS_LAPACK", headers=[SUNDIALS_CONFIG_H],
        include_dirs=include_dirs
    ):
        has_lapack = True
    else:
        has_lapack = False

    cfg = dict(
        float_type = SUNDIALS_FLOAT_TYPE,
        index_size = SUNDIALS_INDEX_SIZE,
        has_lapack = has_lapack,
    )

    return write_pxi(join(BASE_PATH, "sundials_config.pxi"), dict(
        SUNDIALS_FLOAT_TYPE=SUNDIALS_FLOAT_TYPE,
        SUNDIALS_INDEX_SIZE=SUNDIALS_INDEX_SIZE,
        SUNDIALS_BLAS_LAPACK=str(has_lapack),
    )), cfg


class build_ext(_build_ext):
    """
        Custom distutils command which encapsulates api_gen pre-building,
        Cython building, and C compilation.
    """
    def _get_cython_ext(self):
        import numpy
        base_path = 'src/scikits_odes_sundials'
        base_module = "scikits_odes_sundials"

        SUNDIALS_LIBRARIES = []
        CVODE_LIBRARIES = []
        IDA_LIBRARIES = []
        CVODES_LIBRARIES = []
        IDAS_LIBRARIES = []

        SUNDIALS_LIBRARY_DIRS = []
        CVODE_LIBRARY_DIRS = []
        IDA_LIBRARY_DIRS = []
        CVODES_LIBRARY_DIRS = []
        IDAS_LIBRARY_DIRS = []

        # numpy is required for all the extensions
        SUNDIALS_INCLUDE_DIRS = [numpy.get_include()]
        CVODE_INCLUDE_DIRS = []
        IDA_INCLUDE_DIRS = []
        CVODES_INCLUDE_DIRS = []
        IDAS_INCLUDE_DIRS = []

        SUNDIALS_LIBDIR = os.environ.get("SUNDIALS_LIBDIR")
        SUNDIALS_INCLUDEDIR = os.environ.get("SUNDIALS_INCLUDEDIR")
        SUNDIALS_INST_PREFIX = os.environ.get("SUNDIALS_INST")

        if SUNDIALS_LIBDIR or SUNDIALS_INCLUDEDIR:
            SUNDIALS_INCLUDE_DIRS.extend(
                [SUNDIALS_INCLUDEDIR] if SUNDIALS_INCLUDEDIR is not None else []
            )
            SUNDIALS_LIBRARY_DIRS.extend(
                [SUNDIALS_LIBDIR] if SUNDIALS_LIBDIR is not None else []
            )

        elif SUNDIALS_INST_PREFIX is not None:
            SUNDIALS_LIBRARY_DIRS.append(os.path.join(SUNDIALS_INST_PREFIX, "lib"))
            SUNDIALS_INCLUDE_DIRS.append(os.path.join(SUNDIALS_INST_PREFIX, "include"))
            info("SUNDIALS installation path set to `{}` via $SUNDIALS_INST.".format(
                SUNDIALS_INST_PREFIX))
        else:
            info("Searching for SUNDIALS path...")

            # use pkgconfig to find sundials
            try:
                import pkgconfig
                from pkgconfig.pkgconfig import PackageNotFoundError
                try:
                    cvode_pkgconf = pkgconfig.parse(PKGCONFIG_CVODE)
                    for d in cvode_pkgconf.get('library_dirs', []):
                        CVODE_LIBRARY_DIRS.append(str(d))
                    for d in cvode_pkgconf.get('include_dirs', []):
                        CVODE_INCLUDE_DIRS.append(str(d))
                    for lib in cvode_pkgconf.get('include_dirs', []):
                        CVODE_LIBRARIES.append(str(lib))

                    ida_pkgconf = pkgconfig.parse(PKGCONFIG_IDA)
                    for d in ida_pkgconf.get('library_dirs', []):
                        IDA_LIBRARY_DIRS.append(str(d))
                    for d in ida_pkgconf.get('include_dirs', []):
                        IDA_INCLUDE_DIRS.append(str(d))
                    for lib in ida_pkgconf.get('include_dirs', []):
                        IDA_LIBRARIES.append(str(lib))


                    cvodes_pkgconf = pkgconfig.parse(PKGCONFIG_CVODES)
                    for d in cvodes_pkgconf.get('library_dirs', []):
                        CVODES_LIBRARY_DIRS.append(str(d))
                    for d in cvodes_pkgconf.get('include_dirs', []):
                        CVODES_INCLUDE_DIRS.append(str(d))
                    for lib in cvodes_pkgconf.get('include_dirs', []):
                        CVODES_LIBRARIES.append(str(lib))

                    idas_pkgconf = pkgconfig.parse(PKGCONFIG_IDAS)
                    for d in idas_pkgconf.get('library_dirs', []):
                        IDAS_LIBRARY_DIRS.append(str(d))
                    for d in idas_pkgconf.get('include_dirs', []):
                        IDAS_INCLUDE_DIRS.append(str(d))
                    for lib in idas_pkgconf.get('include_dirs', []):
                        IDAS_LIBRARIES.append(str(lib))
                except (EnvironmentError, PackageNotFoundError) as e:
                    pass
            except ImportError:
                info("pkgconfig module not found, using preset paths")

        sundials_pxi, cfg = get_sundials_config_pxi(SUNDIALS_INCLUDE_DIRS,
                self.distribution)

        has_lapack = cfg['has_lapack']

        if not SUNDIALS_LIBRARIES:
            # This is where to put N_vector codes (currently only serial is
            # supported)
            SUNDIALS_LIBRARIES.append('sundials_nvecserial')
            # SUNDIALS_LIBRARIES.append('sundials_nvecopenmp')
            # SUNDIALS_LIBRARIES.append('sundials_nvecparallel')
            # SUNDIALS_LIBRARIES.append('sundials_nvecparhyp')
            # SUNDIALS_LIBRARIES.append('sundials_nvecpetsc')
            # SUNDIALS_LIBRARIES.append('sundials_nvecpthreads')

            # This is where to put SUNLinearSolver codes (klu not supported
            # yet)
            if has_lapack:
                SUNDIALS_LIBRARIES.append('sundials_sunlinsollapackband')
                SUNDIALS_LIBRARIES.append('sundials_sunlinsollapackdense')

            SUNDIALS_LIBRARIES.append('sundials_sunlinsolband')
            SUNDIALS_LIBRARIES.append('sundials_sunlinsoldense')
            SUNDIALS_LIBRARIES.append('sundials_sunlinsolpcg')
            SUNDIALS_LIBRARIES.append('sundials_sunlinsolspbcgs')
            SUNDIALS_LIBRARIES.append('sundials_sunlinsolspfgmr')
            SUNDIALS_LIBRARIES.append('sundials_sunlinsolspgmr')
            SUNDIALS_LIBRARIES.append('sundials_sunlinsolsptfqmr')
            # SUNDIALS_LIBRARIES.append('sundials_sunlinsolklu')

            # This is where to put SUNMatrix codes
            SUNDIALS_LIBRARIES.append('sundials_sunmatrixband')
            SUNDIALS_LIBRARIES.append('sundials_sunmatrixdense')
            SUNDIALS_LIBRARIES.append('sundials_sunmatrixsparse')

        if not IDA_LIBRARIES:
            IDA_LIBRARIES.append('sundials_ida')

        if not CVODE_LIBRARIES:
            CVODE_LIBRARIES.append('sundials_cvode')

        if not IDAS_LIBRARIES:
            IDAS_LIBRARIES.append('sundials_idas')

        if not CVODES_LIBRARIES:
            CVODES_LIBRARIES.append('sundials_cvodes')

        CVODE_LIBRARIES.extend(SUNDIALS_LIBRARIES)
        IDA_LIBRARIES.extend(SUNDIALS_LIBRARIES)
        CVODES_LIBRARIES.extend(SUNDIALS_LIBRARIES)
        IDAS_LIBRARIES.extend(SUNDIALS_LIBRARIES)
        CVODE_INCLUDE_DIRS.extend(SUNDIALS_INCLUDE_DIRS)
        IDA_INCLUDE_DIRS.extend(SUNDIALS_INCLUDE_DIRS)
        CVODES_INCLUDE_DIRS.extend(SUNDIALS_INCLUDE_DIRS)
        IDAS_INCLUDE_DIRS.extend(SUNDIALS_INCLUDE_DIRS)
        CVODE_LIBRARY_DIRS.extend(SUNDIALS_LIBRARY_DIRS)
        IDA_LIBRARY_DIRS.extend(SUNDIALS_LIBRARY_DIRS)
        CVODES_LIBRARY_DIRS.extend(SUNDIALS_LIBRARY_DIRS)
        IDAS_LIBRARY_DIRS.extend(SUNDIALS_LIBRARY_DIRS)

        return [
            Extension(
                base_module + '.' + "common_defs",
                sources = [join(base_path, 'common_defs.pyx')],
                include_dirs=SUNDIALS_INCLUDE_DIRS,
                library_dirs=SUNDIALS_LIBRARY_DIRS,
                libraries=SUNDIALS_LIBRARIES,
            ),
            Extension(
                base_module + '.' + "cvode",
                sources = [join(base_path, 'cvode.pyx')],
                include_dirs=CVODE_INCLUDE_DIRS,
                library_dirs=CVODE_LIBRARY_DIRS,
                libraries=CVODE_LIBRARIES,
            ),
            Extension(
                base_module + '.' + "ida",
                sources = [join(base_path, 'ida.pyx')],
                include_dirs=IDA_INCLUDE_DIRS,
                library_dirs=IDA_LIBRARY_DIRS,
                libraries=IDA_LIBRARIES,
            ),
            Extension(
                base_module + '.' + "cvodes",
                sources = [join(base_path, 'cvodes.pyx')],
                include_dirs=CVODES_INCLUDE_DIRS,
                library_dirs=CVODES_LIBRARY_DIRS,
                libraries=CVODES_LIBRARIES,
            ),
            Extension(
                base_module + '.' + "idas",
                sources = [join(base_path, 'idas.pyx')],
                include_dirs=IDAS_INCLUDE_DIRS,
                library_dirs=IDAS_LIBRARY_DIRS,
                libraries=IDAS_LIBRARIES,
            ),
        ]


    def run(self):
        """ Distutils calls this method to run the command """
        from Cython.Build import cythonize
        self.extensions = cythonize(
            self._get_cython_ext(),
            compiler_directives={
                'language_level': 3,
            },
        )
        _build_ext.run(self) # actually do the build




from .c_sundials cimport *
from libc.stdio cimport FILE

cdef extern from "cvode/cvode.h":
    # lmm
    enum: CV_ADAMS # 1
    enum: CV_BDF   # 2

    # itask
    enum: CV_NORMAL     # 1
    enum: CV_ONE_STEP   # 2

    # CVODE return flags
    enum: CV_SUCCESS           #    0
    enum: CV_TSTOP_RETURN      #    1
    enum: CV_ROOT_RETURN       #    2

    enum: CV_WARNING           #   99

    enum: CV_TOO_MUCH_WORK     #   -1
    enum: CV_TOO_MUCH_ACC      #   -2
    enum: CV_ERR_FAILURE       #   -3
    enum: CV_CONV_FAILURE      #   -4

    enum: CV_LINIT_FAIL        #   -5
    enum: CV_LSETUP_FAIL       #   -6
    enum: CV_LSOLVE_FAIL       #   -7
    enum: CV_RHSFUNC_FAIL      #   -8
    enum: CV_FIRST_RHSFUNC_ERR #   -9
    enum: CV_REPTD_RHSFUNC_ERR #   -10
    enum: CV_UNREC_RHSFUNC_ERR #   -11
    enum: CV_RTFUNC_FAIL       #   -12
    enum: CV_NLS_INIT_FAIL     #   -13
    enum: CV_NLS_SETUP_FAIL    #   -14
    enum: CV_CONSTR_FAIL       #   -15
    enum: CV_NLS_FAIL          #   -16

    enum: CV_MEM_FAIL          #   -20
    enum: CV_MEM_NULL          #   -21
    enum: CV_ILL_INPUT         #   -22
    enum: CV_NO_MALLOC         #   -23
    enum: CV_BAD_K             #   -24
    enum: CV_BAD_T             #   -25
    enum: CV_BAD_DKY           #   -26
    enum: CV_TOO_CLOSE         #   -27
    enum: CV_VECTOROP_ERR      #   -28
    
    enum: CV_UNRECOGNIZED_ERR  #   -99

    ctypedef int (*CVRhsFn)(sunrealtype t, N_Vector y, N_Vector ydot, void *user_data) except? -1
    ctypedef int (*CVRootFn)(sunrealtype t, N_Vector y, sunrealtype *gout, void *user_data) except? -1
    ctypedef int (*CVEwtFn)(N_Vector y, N_Vector ewt, void *user_data)
    ctypedef void (*CVErrHandlerFn)(int error_code,
                               char *module, char *function,
                               char *msg, void *user_data)

    void *CVodeCreate(int lmm, SUNContext sunctx)

    int CVodeInit(void *cvode_mem, CVRhsFn f, sunrealtype t0, N_Vector y0)
    int CVodeReInit(void *cvode_mem, sunrealtype t0, N_Vector y0)

    int CVodeSStolerances(void *cvode_mem, sunrealtype reltol, sunrealtype abstol)
    int CVodeSVtolerances(void *cvode_mem, sunrealtype reltol, N_Vector abstol)
    int CVodeWFtolerances(void *cvode_mem, CVEwtFn efun)

    int CVodeSetErrHandlerFn(void *cvode_mem, CVErrHandlerFn ehfun, void *eh_data)
    int CVodeSetErrFile(void *cvode_mem, FILE *errfp)
    int CVodeSetUserData(void *cvode_mem, void *user_data)
    int CVodeSetMaxOrd(void *cvode_mem, int maxord)
    int CVodeSetMaxNumSteps(void *cvode_mem, long int mxsteps)
    int CVodeSetMaxHnilWarns(void *cvode_mem, int mxhnil)
    int CVodeSetStabLimDet(void *cvode_mem, sunbooleantype stldet)
    int CVodeSetInitStep(void *cvode_mem, sunrealtype hin)
    int CVodeSetMinStep(void *cvode_mem, sunrealtype hmin)
    int CVodeSetMaxStep(void *cvode_mem, sunrealtype hmax)
    int CVodeSetStopTime(void *cvode_mem, sunrealtype tstop)
    int CVodeSetMaxErrTestFails(void *cvode_mem, int maxnef)
    int CVodeSetMaxNonlinIters(void *cvode_mem, int maxcor)
    int CVodeSetMaxConvFails(void *cvode_mem, int maxncf)
    int CVodeSetNonlinConvCoef(void *cvode_mem, sunrealtype nlscoef)
    int CVodeSetConstraints(void *cvode_mem, N_Vector constraints)

    int CVodeSetNonlinearSolver(void *cvode_mem, SUNNonlinearSolver NLS)

    int CVodeRootInit(void *cvode_mem, int nrtfn, CVRootFn g)
    int CVodeSetRootDirection(void *cvode_mem, int *rootdir)
    int CVodeSetNoInactiveRootWarn(void *cvode_mem)

    int CVode(void *cvode_mem, sunrealtype tout, N_Vector yout,
                          sunrealtype *tret, int itask)
    int CVodeGetDky(void *cvode_mem, sunrealtype t, int k, N_Vector dky)

    int CVodeGetWorkSpace(void *cvode_mem, long int *lenrw, long int *leniw)
    int CVodeGetNumSteps(void *cvode_mem, long int *nsteps)
    int CVodeGetNumRhsEvals(void *cvode_mem, long int *nfevals)
    int CVodeGetNumLinSolvSetups(void *cvode_mem, long int *nlinsetups)
    int CVodeGetNumErrTestFails(void *cvode_mem, long int *netfails)
    int CVodeGetLastOrder(void *cvode_mem, int *qlast)
    int CVodeGetCurrentOrder(void *cvode_mem, int *qcur)
    int CVodeGetCurrentGamma(void *cvode_mem, sunrealtype *gamma)
    int CVodeGetNumStabLimOrderReds(void *cvode_mem, long int *nslred)
    int CVodeGetActualInitStep(void *cvode_mem, sunrealtype *hinused)
    int CVodeGetLastStep(void *cvode_mem, sunrealtype *hlast)
    int CVodeGetCurrentStep(void *cvode_mem, sunrealtype *hcur)
    int CVodeGetCurrentState(void *cvode_mem, N_Vector *y)
    int CVodeGetCurrentTime(void *cvode_mem, sunrealtype *tcur)
    int CVodeGetTolScaleFactor(void *cvode_mem, sunrealtype *tolsfac)
    int CVodeGetErrWeights(void *cvode_mem, N_Vector eweight)
    int CVodeGetEstLocalErrors(void *cvode_mem, N_Vector ele)
    int CVodeGetNumGEvals(void *cvode_mem, long int *ngevals)
    int CVodeGetRootInfo(void *cvode_mem, int *rootsfound)

    int CVodeGetIntegratorStats(void *cvode_mem, long int *nsteps,
                                            long int *nfevals, long int *nlinsetups,
                                            long int *netfails, int *qlast,
                                            int *qcur, sunrealtype *hinused, sunrealtype *hlast,
                                            sunrealtype *hcur, sunrealtype *tcur)

    int CVodeGetNumNonlinSolvIters(void *cvode_mem, long int *nniters)
    int CVodeGetNumNonlinSolvConvFails(void *cvode_mem, long int *nncfails)
    int CVodeGetNonlinSolvStats(void *cvode_mem, long int *nniters,
                                            long int *nncfails)
    char *CVodeGetReturnFlagName(long int flag)
    void CVodeFree(void **cvode_mem)

cdef extern from "cvode/cvode_ls.h":
    #CVDLS return values
    enum: CVLS_SUCCESS         #  0
    enum: CVLS_MEM_NULL        # -1
    enum: CVLS_LMEM_NULL       # -2
    enum: CVLS_ILL_INPUT       # -3
    enum: CVLS_MEM_FAIL        # -4
    enum: CVLS_PMEM_NULL        # -5
    enum: CVLS_JACFUNC_UNRECVR  # -6
    enum: CVLS_JACFUNC_RECVR    # -7
    enum: CVLS_SUNMAT_FAIL      # -8
    enum: CVLS_SUNLS_FAIL       # -9

    ctypedef int (*CVLsJacFn)(sunrealtype t, N_Vector y, N_Vector fy,
                              SUNMatrix Jac, void *user_data,
                              N_Vector tmp1, N_Vector tmp2, N_Vector tmp3) except? -1

    ctypedef int (*CVLsPrecSetupFn)(sunrealtype t, N_Vector y, N_Vector fy,
                                    sunbooleantype jok, sunbooleantype *jcurPtr,
                                    sunrealtype gamma, void *user_data) except? -1

    ctypedef int (*CVLsPrecSolveFn)(sunrealtype t, N_Vector y, N_Vector fy,
                                    N_Vector r, N_Vector z, sunrealtype gamma,
                                    sunrealtype delta, int lr, void *user_data) except? -1

    ctypedef int (*CVLsJacTimesSetupFn)(sunrealtype t, N_Vector y,
                                       N_Vector fy, void *user_data) except? -1

    ctypedef int (*CVLsJacTimesVecFn)(N_Vector v, N_Vector Jv, sunrealtype t,
                                      N_Vector y, N_Vector fy,
                                      void *user_data, N_Vector tmp) except? -1

    ctypedef int (*CVLsLinSysFn)(sunrealtype t, N_Vector y, N_Vector fy, 
                                 SUNMatrix A, sunbooleantype jok, 
                                 sunbooleantype *jcur, sunrealtype gamma,
                                 void *user_data, N_Vector tmp1, N_Vector tmp2,
                                 N_Vector tmp3)


    int CVodeSetLinearSolver(void *cvode_mem, SUNLinearSolver LS, SUNMatrix A)

    int CVodeSetJacFn(void *cvode_mem, CVLsJacFn jac)
    int CVodeSetMaxStepsBetweenJac(void *cvode_mem, long int msbj)
    int CVodeSetEpsLin(void *cvode_mem, sunrealtype eplifac)
    int CVodeSetPreconditioner(void *cvode_mem, CVLsPrecSetupFn pset,
                               CVLsPrecSolveFn psolve)
    int CVodeSetJacTimes(void *cvode_mem, CVLsJacTimesSetupFn jtsetup,
                         CVLsJacTimesVecFn jtimes)
    int CVodeSetLinSysFn(void *cvode_mem, CVLsLinSysFn linsys)

    int CVodeGetLinWorkSpace(void *cvode_mem, long int *lenrwLS,
                             long int *leniwLS)
    int CVodeGetNumJacEvals(void *cvode_mem, long int *njevals)
    int CVodeGetNumPrecEvals(void *cvode_mem, long int *npevals)
    int CVodeGetNumPrecSolves(void *cvode_mem, long int *npsolves)
    int CVodeGetNumLinIters(void *cvode_mem, long int *nliters)
    int CVodeGetNumLinConvFails(void *cvode_mem, long int *nlcfails)
    int CVodeGetNumJTSetupEvals(void *cvode_mem, long int *njtsetups)
    int CVodeGetNumJtimesEvals(void *cvode_mem, long int *njvevals)
    int CVodeGetNumLinRhsEvals(void *cvode_mem, long int *nfevalsLS)
    int CVodeGetLastLinFlag(void *cvode_mem, long int *flag)
    char *CVodeGetLinReturnFlagName(long int flag)

cdef extern from "cvode/cvode_direct.h":
    ctypedef CVLsJacFn CVodeJacFn

    int CVodeSetJacFn(void *cvode_mem, CVodeJacFn jac)
    int CVodeGetWorkSpace(void *cvode_mem, long int *lenrwLS, long int *leniwLS)
    int CVodeGetNumJacEvals(void *cvode_mem, long int *njevals)
    int CVodeGetNumRhsEvals(void *cvode_mem, long int *nfevalsLS)
    int CVodeGetLastFlag(void *cvode_mem, long int *flag)
    char *CVodeGetReturnFlagName(long int flag)

cdef extern from "cvode/cvode_bandpre.h":
    int CVBandPrecInit(void *cvode_mem, sunindextype N, sunindextype mu,
                       sunindextype ml);
    int CVBandPrecGetWorkSpace(void *cvode_mem, long int *lenrwLS, 
                               long int *leniwLS)
    int CVBandPrecGetNumRhsEvals(void *cvode_mem, long int *nfevalsBP)

cdef extern from "cvode/cvode_diag.h":
    # CVDIAG return values
    enum: CVDIAG_SUCCESS         #  0
    enum: CVDIAG_MEM_NULL        # -1
    enum: CVDIAG_LMEM_NULL       # -2
    enum: CVDIAG_ILL_INPUT       # -3
    enum: CVDIAG_MEM_FAIL        # -4
    # Additional last_flag values 
    enum: CVDIAG_INV_FAIL        # -5
    enum: CVDIAG_RHSFUNC_UNRECVR # -6
    enum: CVDIAG_RHSFUNC_RECVR   # -7

    int CVDiag(void *cvode_mem)
    int CVDiagGetWorkSpace(void *cvode_mem, long int *lenrwLS, long int *leniwLS)
    int CVDiagGetNumRhsEvals(void *cvode_mem, long int *nfevalsLS)
    int CVDiagGetLastFlag(void *cvode_mem, long int *flag)
    char *CVDiagGetReturnFlagName(long int flag)

cdef extern from "cvode/cvode_bbdpre.h":
    ctypedef int (*CVLocalFn)(sunindextype Nlocal, sunrealtype t, N_Vector y,
                              N_Vector g, void *user_data)
    ctypedef int (*CVCommFn)(sunindextype Nlocal, sunrealtype t, N_Vector y,
                             void *user_data)

    int CVBBDPrecInit(void *cvode_mem, sunindextype Nlocal,
                      sunindextype mudq, sunindextype mldq,
                      sunindextype mukeep, sunindextype mlkeep,
                      sunrealtype dqrely, CVLocalFn gloc, CVCommFn cfn)
    int CVBBDPrecReInit(void *cvode_mem, sunindextype mudq, sunindextype mldq,
                        sunrealtype dqrely)
    int CVBBDPrecGetWorkSpace(void *cvode_mem, long int *lenrwBBDP, 
                              long int *leniwBBDP)
    int CVBBDPrecGetNumGfnEvals(void *cvode_mem, long int *ngevalsBBDP)

cdef extern from "cvode/cvode_spils.h":

    ctypedef CVLsPrecSetupFn CVodePrecSetupFn
    ctypedef CVLsPrecSolveFn CVodePrecSolveFn
    ctypedef CVLsJacTimesSetupFn CVodeJacTimesSetupFn
    ctypedef CVLsJacTimesVecFn CVodeJacTimesVecFn

    int CVodeSetEpsLin(void *cvode_mem, sunrealtype eplifac)
    int CVodeSetPreconditioner(void *cvode_mem, CVodePrecSetupFn pset,
                                 CVodePrecSolveFn psolve)
    int CVodeSetJacTimes(void *cvode_mem, CVodeJacTimesSetupFn jtsetup,
                           CVodeJacTimesVecFn jtimes)

    int CVodeGetWorkSpace(void *cvode_mem, long int *lenrwLS, long int *leniwLS)
    int CVodeGetNumPrecEvals(void *cvode_mem, long int *npevals)
    int CVodeGetNumPrecSolves(void *cvode_mem, long int *npsolves)
    int CVodeGetNumLinIters(void *cvode_mem, long int *nliters)
    int CVodeGetNumConvFails(void *cvode_mem, long int *nlcfails)
    int CVodeGetNumJTSetupEvals(void *cvode_mem, long int *njtsetups)
    int CVodeGetNumJtimesEvals(void *cvode_mem, long int *njvevals)
    int CVodeGetNumRhsEvals(void *cvode_mem, long int *nfevalsLS)
    int CVodeGetLastFlag(void *cvode_mem, long int *flag)
    char *CVodeGetReturnFlagName(long int flag)



"""
SUNDIALS wrapper
"""

import inspect
from . import _version
__version__ = _version.get_versions()['version']


class CVODESolveException(Exception):
    """Base class for exceptions raised by ``CVODE.validate_flags``."""
    def __init__(self, soln):
        self.soln = soln
        self.args = (self._message.format(soln),)

class CVODESolveFailed(CVODESolveException):
    """``CVODE.solve`` failed to reach endpoint"""
    _message = (
        "Solver failed with flag {0.flag} and finished at {0.errors.t}"
        "with values {0.errors.y}."
    )

class CVODESolveFoundRoot(CVODESolveException):
    """``CVODE.solve`` found a root"""
    _message = "Solver found a root at {0.roots.t[0]}."

class CVODESolveReachedTSTOP(CVODESolveException):
    """``CVODE.solve`` reached the endpoint specified by tstop."""
    _message = "Solver reached tstop at {0.tstop.t[0]}."

class IDASolveException(Exception):
    """Base class for exceptions raised by ``IDA.validate_flags``."""
    def __init__(self, soln):
        self.soln = soln
        self.args = (self._message.format(soln),)

class IDASolveFailed(IDASolveException):
    """``IDA.solve`` failed to reach endpoint"""
    _message = (
        "Solver failed with flag {0.flag} and finished at {0.errors.t}"
        "with values {0.errors.y} and derivatives {0.errors.ydot}."
    )

class IDASolveFoundRoot(IDASolveException):
    """``IDA.solve`` found a root"""
    _message = "Solver found a root at {0.roots.t[0]}."

class IDASolveReachedTSTOP(IDASolveException):
    """``IDA.solve`` reached the endpoint specified by tstop."""
    _message = "Solver reached tstop at {0.tstop.t[0]}."


def _get_num_args(func):
    """
    Python 2/3 compatible method of getting number of args that `func` accepts
    """
    if hasattr(inspect, "getfullargspec"):
        argspec = inspect.getfullargspec(func)
    else:
        argspec = inspect.getargspec(func)
    arg_cnt = 0
    for arg in argspec.args:
        if arg not in ("self", "cls"):
            arg_cnt += 1
    return arg_cnt



cimport numpy as np
from .c_sundials cimport N_Vector, SUNDlsMat, SUNMatrix

include "sundials_config.pxi"

IF SUNDIALS_FLOAT_TYPE == "single":
    ctypedef np.float_t DTYPE_t
ELIF SUNDIALS_FLOAT_TYPE == "double":
    ctypedef np.double_t DTYPE_t
ELIF SUNDIALS_FLOAT_TYPE == "extended":
    ctypedef np.longdouble_t DTYPE_t
ELSE:
    # fall back to double
    ctypedef np.double_t DTYPE_t

IF SUNDIALS_INDEX_SIZE == "32":
    ctypedef np.int32_t INDEX_TYPE_t
ELIF SUNDIALS_INDEX_SIZE == "64":
    ctypedef np.int64_t INDEX_TYPE_t
ELSE:
    ctypedef np.int64_t INDEX_TYPE_t

cdef int nv_s2ndarray(N_Vector v, np.ndarray[DTYPE_t, ndim=1] a) except? -1
cdef int ndarray2nv_s(N_Vector v, np.ndarray[DTYPE_t, ndim=1] a) except? -1

cdef int SUNMatrix2ndarray(SUNMatrix m, np.ndarray a) except? -1
cdef int ndarray2SUNMatrix(SUNMatrix m, np.ndarray a) except? -1

cdef ensure_numpy_float_array(object value)



from libc.stdio cimport FILE
from .c_sundials cimport *

cdef extern from "nvector/nvector_serial.h":
    cdef struct _N_VectorContent_Serial:
        sunindextype length
        sunbooleantype own_data
        sunrealtype *data

    ctypedef _N_VectorContent_Serial *N_VectorContent_Serial

    N_Vector N_VNew_Serial(sunindextype vec_length, SUNContext sunctx)
    N_Vector N_VNewEmpty_Serial(sunindextype vec_length, SUNContext sunctx)
    N_Vector N_VMake_Serial(sunindextype vec_length, sunrealtype *v_data, SUNContext sunctx)
    N_Vector *N_VCloneVectorArray_Serial(int count, N_Vector w)
    N_Vector *N_VCloneVectorArrayEmpty_Serial(int count, N_Vector w)
    void N_VDestroyVectorArray_Serial(N_Vector *vs, int count)
    sunindextype N_VGetLength_Serial(N_Vector v)
    void N_VPrint_Serial(N_Vector v)
    void N_VPrintFile_Serial(N_Vector v, FILE *outfile)

    N_Vector_ID N_VGetVectorID_Serial(N_Vector v)
    N_Vector N_VCloneEmpty_Serial(N_Vector w)
    N_Vector N_VClone_Serial(N_Vector w)
    void N_VDestroy_Serial(N_Vector v)
    void N_VSpace_Serial(N_Vector v, sunindextype *lrw, sunindextype *liw)
    sunrealtype *N_VGetArrayPointer_Serial(N_Vector v)
    void N_VSetArrayPointer_Serial(sunrealtype *v_data, N_Vector v)
    void N_VLinearSum_Serial(sunrealtype a, N_Vector x, sunrealtype b, N_Vector y, N_Vector z)
    void N_VConst_Serial(sunrealtype c, N_Vector z)
    void N_VProd_Serial(N_Vector x, N_Vector y, N_Vector z)
    void N_VDiv_Serial(N_Vector x, N_Vector y, N_Vector z)
    void N_VScale_Serial(sunrealtype c, N_Vector x, N_Vector z)
    void N_VAbs_Serial(N_Vector x, N_Vector z)
    void N_VInv_Serial(N_Vector x, N_Vector z)
    void N_VAddConst_Serial(N_Vector x, sunrealtype b, N_Vector z)
    sunrealtype N_VDotProd_Serial(N_Vector x, N_Vector y)
    sunrealtype N_VMaxNorm_Serial(N_Vector x)
    sunrealtype N_VWrmsNorm_Serial(N_Vector x, N_Vector w)
    sunrealtype N_VWrmsNormMask_Serial(N_Vector x, N_Vector w, N_Vector id)
    sunrealtype N_VMin_Serial(N_Vector x)
    sunrealtype N_VWL2Norm_Serial(N_Vector x, N_Vector w)
    sunrealtype N_VL1Norm_Serial(N_Vector x)
    void N_VCompare_Serial(sunrealtype c, N_Vector x, N_Vector z)
    sunbooleantype N_VInvTest_Serial(N_Vector x, N_Vector z)
    sunbooleantype N_VConstrMask_Serial(N_Vector c, N_Vector x, N_Vector m)
    sunrealtype N_VMinQuotient_Serial(N_Vector num, N_Vector denom)
    
    
    int N_VLinearCombination_Serial(int nvec, sunrealtype* c, N_Vector* V,
                                    N_Vector z)
    int N_VScaleAddMulti_Serial(int nvec, sunrealtype* a, N_Vector x,
                                N_Vector* Y, N_Vector* Z)
    int N_VDotProdMulti_Serial(int nvec, N_Vector x, N_Vector* Y, 
                               sunrealtype* dotprods)

    int N_VLinearSumVectorArray_Serial(int nvec, sunrealtype a, N_Vector* X,
                                       sunrealtype b, N_Vector* Y, N_Vector* Z)
    int N_VScaleVectorArray_Serial(int nvec, sunrealtype* c,
                                   N_Vector* X, N_Vector* Z)
    int N_VConstVectorArray_Serial(int nvecs, sunrealtype c, N_Vector* Z)
    int N_VWrmsNormVectorArray_Serial(int nvecs, N_Vector* X,
                                      N_Vector* W, sunrealtype* nrm)
    int N_VWrmsNormMaskVectorArray_Serial(int nvecs, N_Vector* X,
                                          N_Vector* W, N_Vector id,
                                          sunrealtype* nrm)
    int N_VScaleAddMultiVectorArray_Serial(int nvec, int nsum, sunrealtype* a,
                                           N_Vector* X, N_Vector** Y,
                                           N_Vector** Z)
    int N_VLinearCombinationVectorArray_Serial(int nvec, int nsum,
                                               sunrealtype* c,
                                               N_Vector** X, N_Vector* Z)

    sunrealtype N_VWSqrSumLocal_Serial(N_Vector x, N_Vector w)
    sunrealtype N_VWSqrSumMaskLocal_Serial(N_Vector x, N_Vector w, N_Vector id)
  
    int N_VEnableFusedOps_Serial(N_Vector v, sunbooleantype tf)

    int N_VEnableLinearCombination_Serial(N_Vector v, sunbooleantype tf)
    int N_VEnableScaleAddMulti_Serial(N_Vector v, sunbooleantype tf)
    int N_VEnableDotProdMulti_Serial(N_Vector v, sunbooleantype tf)

    int N_VEnableLinearSumVectorArray_Serial(N_Vector v, sunbooleantype tf)
    int N_VEnableScaleVectorArray_Serial(N_Vector v, sunbooleantype tf)
    int N_VEnableConstVectorArray_Serial(N_Vector v, sunbooleantype tf)
    int N_VEnableWrmsNormVectorArray_Serial(N_Vector v, sunbooleantype tf)
    int N_VEnableWrmsNormMaskVectorArray_Serial(N_Vector v, sunbooleantype tf)
    int N_VEnableScaleAddMultiVectorArray_Serial(N_Vector v, sunbooleantype tf)
    int N_VEnableLinearCombinationVectorArray_Serial(N_Vector v, sunbooleantype tf)

    # Macros
    sunindextype NV_LENGTH_S(N_Vector vc_s)
    sunrealtype* NV_DATA_S(N_Vector vc_s)

